"""
MQTTk - Lightweight graphical MQTT client and message analyser

Copyright (C) 2022  Máté Szabó

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
import os.path
import traceback

from mqttk import __version__ as version
import tkinter as tk
import tkinter.ttk as ttk
from tkinter import filedialog
import json
from mqttk.helpers import validate_name, clear_combobox_selection, get_clear_combobox_selection_function
from tkinter import messagebox
from copy import deepcopy


about_text = "MQTTk is a lightweight, free and open source graphical MQTT\n" \
             "client/analyser. It is licensed under the GNU GPLv3 license.\n" \
             "For license information please see:\n" \
             "https://www.gnu.org/licenses/\n\n" \
             "This software is written in pure Python and is powered by\n" \
             "the below open source projects:\n\n" \
             "Python 3 and Tkinter/ttk:\n" \
             "https://docs.python.org/3/license.html\n\n" \
             "Eclipse paho-mqtt python client library:\n" \
             "https://github.com/eclipse/paho.mqtt.python\n\n" \
             "xmltodict library:\n" \
             "https://github.com/martinblech/xmltodict\n\n" \
             "Copyright (C) 2022  Máté Szabó"


class AboutDialog(tk.Toplevel):
    def __init__(self, master, icon, style):
        super().__init__(master=master)
        self.transient(master)
        self.master = master
        self.title("About")

        self.resizable(False, False)
        self.iconphoto(False, icon)

        self.about_frame = ttk.Frame(self)

        self.about_label = ttk.Label(self.about_frame, text="About MQTTk\n\nVersion {}".format(version), anchor='n', justify=tk.CENTER, font="Arial 14 bold")
        self.about_label.pack(side=tk.TOP, fill='x', expand=1)

        self.about_content_frame = ttk.Frame(self.about_frame)
        # self.icon_canvas = tk.Canvas(self.about_content_frame, width=200, height=200)
        # self.icon_canvas.create_image(10, 10, anchor='w', image=icon)
        # self.icon_canvas.pack(side=tk.LEFT, expand=1, fill='both')
        self.icon = ttk.Label(self.about_content_frame, image=icon)
        self.icon.pack(side=tk.LEFT, padx=8)
        self.about_text = tk.Text(self.about_content_frame, wrap='word', width=55, height=18, exportselection=False, font='Arial 14')
        self.about_text.pack(side=tk.RIGHT, fill='x', expand=1, padx=6, pady=6)
        self.about_text.insert(1.0, about_text)
        self.about_text.configure(bg=style.lookup("TLabel", "background"),
                                  relief='flat',
                                  state='disabled',
                                  borderwidth=0,
                                  border=0,
                                  highlightcolor=style.lookup("TLabel", "background"))
        self.about_content_frame.pack(expand=1, fill='x')

        self.ok_button = ttk.Button(self.about_frame, text="OK")
        self.ok_button.pack(side=tk.BOTTOM, pady=4)
        self.ok_button["command"] = self.on_destroy
        self.about_frame.pack(fill='both', expand=1)

        self.grab_set()

        self.geometry("")
        self.update()
        screenwidth = self.winfo_screenwidth()
        screenheight = self.winfo_screenheight()
        height = self.winfo_height()
        width = self.winfo_width()
        alignstr = '%dx%d+%d+%d' % (width, height, (screenwidth - width) / 2, (screenheight - height) / 2)
        self.geometry(alignstr)
        self.protocol("WM_DELETE_WINDOW", self.on_destroy)
        self.bind("<Return>", self.on_destroy)
        self.bind("<Escape>", self.on_destroy)
        self.focus_set()
        self.wait_window(self)

    def on_destroy(self, *args, **kwargs):
        self.grab_release()
        self.destroy()


class PublishNameDialog(tk.Toplevel):
    def __init__(self, master, current_value, name_callback):
        super().__init__(master=master)
        self.master = master
        self.title("Please enter publish template name")
        self.name_callback = name_callback
        width = 250
        height = 140
        screenwidth = self.winfo_screenwidth()
        screenheight = self.winfo_screenheight()
        alignstr = '%dx%d+%d+%d' % (width, height, (screenwidth - width) / 2, (screenheight - height) / 2)
        self.resizable(False, False)
        self.geometry(alignstr)
        self.protocol("WM_DELETE_WINDOW", self.on_destroy)

        self.dialog_frame = ttk.Frame(self)
        self.dialog_frame.pack(fill='both', expand=1)

        self.instruction_text = ttk.Label(self.dialog_frame, text="Please enter publish template name".format(version), anchor='n', justify=tk.CENTER)
        self.instruction_text.pack(side=tk.TOP, fill='x', padx=10, pady=10)

        self.name_input = ttk.Entry(self.dialog_frame)
        self.name_input.insert(0, current_value)
        self.name_input.pack(expand=1, fill='x', side=tk.TOP, padx=10, pady=10)

        self.buttons_frame = ttk.Frame(self.dialog_frame)
        self.buttons_frame.pack(side=tk.TOP, fill='x', expand=1)
        self.ok_button = ttk.Button(self.dialog_frame, text="OK")
        self.ok_button.pack(side=tk.RIGHT, pady=10, padx=20)
        self.ok_button["command"] = self.on_save
        self.cancel_button = ttk.Button(self.dialog_frame, text="Cancel")
        self.cancel_button.pack(side=tk.LEFT, pady=10, padx=20)
        self.cancel_button["command"] = self.on_destroy

    def on_save(self, ):
        if self.name_input.get() != "":
            self.name_callback(self.name_input.get())
        self.on_destroy()

    def on_destroy(self, *args, **kwargs):
        self.grab_release()
        self.destroy()


class SplashScreen(tk.Toplevel):
    def __init__(self, master, splash_icon):
        super().__init__(master=master)
        screenwidth = master.winfo_screenwidth()
        screenheight = master.winfo_screenheight()
        self.overrideredirect(True)
        width = 370
        height = 350
        alignstr = '%dx%d+%d+%d' % (width, height, (screenwidth - width) / 2, (screenheight - height) / 2)
        self.geometry(alignstr)
        self.title("Splash")
        splash_label = ttk.Label(self, image=splash_icon, anchor=tk.CENTER)
        splash_label.pack(expand=1, fill="both")
        self.update()


class ConnectionConfigImportExport(tk.Toplevel):
    def __init__(self, master, icon, config_handler, logger, is_import=False):
        super().__init__(master=master)
        self.transient(master)
        self.is_import = is_import
        self.master = master
        self.title("{} connection config".format("Import" if is_import else "Export"))
        self.resizable(False, False)
        self.iconphoto(False, icon)
        self.config_handler = config_handler
        self.log = logger
        self.imported_connection_configs = None

        self.dialog_frame = ttk.Frame(self)
        self.dialog_frame.pack(fill="both", expand=1)

        if not is_import:
            self.select_profile_frame = ttk.Frame(self.dialog_frame)
            self.select_profile_frame.pack(expand=1, fill='y', padx=4, pady=4)

            connection_selector_label = ttk.Label(self.select_profile_frame, text="Select connection")
            connection_selector_label.pack(side=tk.LEFT)

            self.connection_selector = ttk.Combobox(self.select_profile_frame, width=30, exportselection=False)
            self.connection_selector.bind("<<ComboboxSelected>>", get_clear_combobox_selection_function(self.connection_selector))
            self.connection_selector.pack(side=tk.LEFT, padx=3, pady=3)
            connection_profile_list = sorted(self.config_handler.get_connection_profiles())
            if len(connection_profile_list) != 0:
                self.connection_selector.configure(values=connection_profile_list)
                if self.config_handler.get_last_used_connection() in connection_profile_list:
                    self.connection_selector.current(
                        connection_profile_list.index(self.config_handler.get_last_used_connection()))
                else:
                    self.connection_selector.current(0)

        self.browse_frame = ttk.Frame(self.dialog_frame)
        self.browse_frame.pack(fill="x", padx=4, pady=4)
        self.browse_label = ttk.Label(self.browse_frame, text="File")
        self.browse_label.pack(side=tk.LEFT, anchor="w", padx=2, pady=4, expand=1)
        self.file_input = ttk.Entry(self.browse_frame, width=40)
        self.file_input.pack(side=tk.LEFT, padx=2)
        self.browser_button = ttk.Button(self.browse_frame, width=3, text="...", command=self.browse_file)
        self.browser_button.pack(side=tk.LEFT, padx=2, pady=2)

        self.info_frame = ttk.Frame(self.dialog_frame)
        self.info_frame.pack(padx=4, pady=4, fill="x", expand=1)
        self.info_label = ttk.Label(self.info_frame)
        self.info_label.pack(fill="x")

        self.button_frame = ttk.Frame(self.dialog_frame)
        self.button_frame.pack(expand=1, pady=4, padx=4, fill="x")
        self.ok_button = ttk.Button(self.button_frame,
                                    text="Import" if is_import else "Export",
                                    command=self.ok,
                                    state="disabled")
        self.ok_button.pack(side=tk.RIGHT)

        self.cancel_button = ttk.Button(self.button_frame,
                                        text="Cancel",
                                        command=self.on_destroy)
        self.cancel_button.pack(side=tk.RIGHT, padx=4, pady=4)

        self.grab_set()
        self.geometry("")
        self.update()
        screenwidth = self.winfo_screenwidth()
        screenheight = self.winfo_screenheight()
        height = self.winfo_height()
        width = self.winfo_width()
        alignstr = '%dx%d+%d+%d' % (width, height, (screenwidth - width) / 2, (screenheight - height) / 2)
        self.geometry(alignstr)
        self.protocol("WM_DELETE_WINDOW", self.on_destroy)
        self.bind("<Escape>", self.on_destroy)
        self.focus_set()
        self.wait_window(self)

    def browse_file(self):
        self.imported_connection_configs = None
        self.ok_button["state"] = "disabled"
        if self.is_import:
            file_path_name = filedialog.askopenfilename(initialdir=self.config_handler.get_last_used_directory(),
                                                        title="Import connection configuration")
            if file_path_name == "":
                self.log.warning("Empty file name when browsing for connection config to import. Maybe the cancel button was pressed?")
                return
        else:
            file_path_name = filedialog.asksaveasfilename(initialdir=self.config_handler.get_last_used_directory(),
                                                          title="Export connection configuration",
                                                          defaultextension="json",
                                                          initialfile="{}-config".format(self.connection_selector.get().replace(" ", "_")))

            if file_path_name == "":
                self.log.warning("Empty file name when exporting connection config. Maybe the cancel button was pressed?")
                return

        self.config_handler.save_last_used_directory(file_path_name)
        self.file_input.delete(0, tk.END)
        self.file_input.insert(0, file_path_name)

        if not self.is_import:
            self.log.info("File selected for communication config export: {}".format(file_path_name))
            self.ok_button["state"] = "normal"
            return

        if not os.path.isfile(file_path_name):
            self.log.error("The file picked for connection config import doesn't exist!", file_path_name)
            self.info_label["text"] = "Invalid file"
            self.info_label["foreground"] = "red"
            return

        try:
            with open(file_path_name, "r", encoding="utf-8") as configfile:
                config_dict = json.loads(configfile.read())
        except Exception as e:
            self.log.error("Failed to load connection config file", e, traceback.format_exc())

        connection_configs = config_dict.get("connections", None)
        if not connection_configs:
            self.log.error("There are no connection profiles in this file.")
            self.info_label["text"] = "There are no connection profiles in this file."
            self.info_label["foreground"] = "red"
            return

        self.info_label["text"] = "Found {} connection configuration{}!".format(len(connection_configs),
                                                                                "" if len(connection_configs) == 1 else "s")
        self.info_label["foreground"] = "green"
        self.imported_connection_configs = connection_configs
        self.ok_button["state"] = "normal"
        self.log.info("File selected for communication config import: {}".format(file_path_name))

    def on_destroy(self, *args, **kwargs):
        self.grab_release()
        self.destroy()

    def ok(self, *args, **kwargs):
        if self.is_import:
            self.log.info("Attempting to import connection configuration")
            try:
                all_connection_configs = self.config_handler.get_connection_profiles()
                for connection_name, configuration in self.imported_connection_configs.items():
                    name = validate_name(connection_name, all_connection_configs)
                    self.config_handler.save_connection_config(name, configuration.get("connection_parameters", {}))
            except Exception as e:
                self.log.error("Failed to import connection configurations", e, traceback.format_exc())
                messagebox.showerror("Error importing communication config", "See log for details")
            else:
                messagebox.showinfo("Success", "Communication configuration{} imported successfully".format(
                    "" if len(self.imported_connection_configs) == 1 else "s"
                ))
            self.on_destroy()
        else:
            connection_to_export = self.connection_selector.get()
            connection_config = self.config_handler.get_connection_config_dict(connection_to_export)
            self.log.info("Exporting connection config {} into {}".format(connection_to_export,
                                                                           self.file_input.get()))
            export_dict = {
                "connections": {
                    connection_to_export: deepcopy(connection_config)
                }
            }

            export_dict["connections"][connection_to_export].pop("subscriptions", None)
            export_dict["connections"][connection_to_export].pop("publish_topics", None)
            export_dict["connections"][connection_to_export].pop("stored_publishes", None)
            export_dict["connections"][connection_to_export].pop("last_publish_used", None)
            export_dict["connections"][connection_to_export].pop("last_subscribe_used", None)

            try:
                with open(self.file_input.get(), "w", encoding="utf-8") as export_file:
                    export_file.write(json.dumps(export_dict, indent=2, ensure_ascii=False))
            except Exception as e:
                self.log.error("Failed to export connection config", e, traceback.format_exc())
                messagebox.showerror("Error", "Error exporting communication config. SSee log for details")
            else:
                self.log.info("Successfully exported connection config")
                messagebox.showinfo("Success", "Communication profile exported successfully")
            self.on_destroy()


class SubscribePublishImportExport(tk.Toplevel):
    def __init__(self, master, icon, config_handler, logger, is_import=False):
        super().__init__(master=master)
        self.transient(master)
        self.is_import = is_import
        self.master = master
        self.title("{} subscribe/publish/template content".format("Import" if is_import else "Export"))
        self.resizable(False, False)
        self.iconphoto(False, icon)
        self.config_handler = config_handler
        self.log = logger
        self.imported_history = None

        self.dialog_frame = ttk.Frame(self)
        self.dialog_frame.pack(fill="both", expand=1)

        self.select_profile_frame = ttk.Frame(self.dialog_frame)
        self.select_profile_frame.pack(expand=1, fill='y', padx=4, pady=4)

        connection_selector_label = ttk.Label(self.select_profile_frame, text="Select connection")
        connection_selector_label.pack(side=tk.LEFT)

        self.connection_selector = ttk.Combobox(self.select_profile_frame, width=30, exportselection=False)
        self.connection_selector.bind("<<ComboboxSelected>>", self.on_profile_select)
        self.connection_selector.pack(side=tk.LEFT, padx=3, pady=3)

        self.browse_frame = ttk.Frame(self.dialog_frame)
        self.browse_frame.pack(fill="x", padx=4, pady=4)
        self.browse_label = ttk.Label(self.browse_frame, text="File")
        self.browse_label.pack(side=tk.LEFT, anchor="w", padx=2, pady=4, expand=1)
        self.file_input = ttk.Entry(self.browse_frame, width=40)
        self.file_input.pack(side=tk.LEFT, padx=2)
        self.browser_button = ttk.Button(self.browse_frame, width=3, text="...", command=self.browse_file)
        self.browser_button.pack(side=tk.LEFT, padx=2, pady=2)

        self.checkbox_frame = ttk.Frame(self.dialog_frame)
        self.checkbox_frame.pack(padx=4, pady=4, expand=1)

        self.subscribe_topic_selection = tk.IntVar()
        self.subscribe_topic_checkbox = ttk.Checkbutton(self.checkbox_frame,
                                                        text="Subscribe topic history items",
                                                        variable=self.subscribe_topic_selection,
                                                        offvalue=0,
                                                        onvalue=1,
                                                        command=self.on_checkbox)
        self.subscribe_topic_checkbox.pack(expand=1, fill="x", padx=4, pady=4)

        self.publish_topic_selection = tk.IntVar()
        self.publish_topic_checkbox = ttk.Checkbutton(self.checkbox_frame,
                                                      text="Publish topic history items",
                                                      variable=self.publish_topic_selection,
                                                      offvalue=0,
                                                      onvalue=1,
                                                      command=self.on_checkbox)
        self.publish_topic_checkbox.pack(expand=1, fill="x", padx=4, pady=4)

        self.message_template_selection = tk.IntVar()
        self.message_template_checkbox = ttk.Checkbutton(self.checkbox_frame,
                                                         text="Message template items",
                                                         variable=self.message_template_selection,
                                                         offvalue=0,
                                                         onvalue=1,
                                                         command=self.on_checkbox)
        self.message_template_checkbox.pack(expand=1, fill="x", padx=4, pady=4)

        self.info_frame = ttk.Frame(self.dialog_frame)
        self.info_frame.pack(padx=4, pady=4, fill="x", expand=1)
        self.info_label = ttk.Label(self.info_frame)
        self.info_label.pack(fill="x")

        self.button_frame = ttk.Frame(self.dialog_frame)
        self.button_frame.pack(expand=1, pady=4, padx=4, fill="x")
        self.ok_button = ttk.Button(self.button_frame,
                                    text="Import" if is_import else "Export",
                                    command=self.ok,
                                    state="disabled")
        self.ok_button.pack(side=tk.RIGHT)

        self.cancel_button = ttk.Button(self.button_frame,
                                        text="Cancel",
                                        command=self.on_destroy)
        self.cancel_button.pack(side=tk.RIGHT, padx=4, pady=4)

        connection_profile_list = sorted(self.config_handler.get_connection_profiles())
        if len(connection_profile_list) != 0:
            self.connection_selector.configure(values=connection_profile_list)
            if self.config_handler.get_last_used_connection() in connection_profile_list:
                self.connection_selector.current(
                    connection_profile_list.index(self.config_handler.get_last_used_connection()))
                self.on_profile_select()
            else:
                self.connection_selector.current(0)
                self.on_profile_select()

        self.grab_set()
        self.geometry("")
        self.update()
        screenwidth = self.winfo_screenwidth()
        screenheight = self.winfo_screenheight()
        height = self.winfo_height()
        width = self.winfo_width()
        alignstr = '%dx%d+%d+%d' % (width, height, (screenwidth - width) / 2, (screenheight - height) / 2)
        self.geometry(alignstr)
        self.protocol("WM_DELETE_WINDOW", self.on_destroy)
        self.bind("<Escape>", self.on_destroy)
        self.focus_set()
        self.wait_window(self)

    def on_checkbox(self):
        if any((bool(self.subscribe_topic_selection.get()),
                bool(self.publish_topic_selection.get()),
                bool(self.message_template_selection.get()))) and self.file_input.get() != "":
            self.ok_button["state"] = "normal"
        else:
            self.ok_button["state"] = "disabled"

    def on_profile_select(self, *args, **kwargs):
        clear_combobox_selection(combobox_instance=self.connection_selector)
        if self.is_import:
            return

        connection_config = self.config_handler.get_connection_config_dict(self.connection_selector.get())
        self.update_items(connection_config)

    def browse_file(self):
        self.imported_history = None
        self.ok_button["state"] = "disabled"
        self.info_label["text"] = ""

        if self.is_import:
            file_path_name = filedialog.askopenfilename(initialdir=self.config_handler.get_last_used_directory(),
                                                        title="Import connection configuration")
            if file_path_name == "":
                self.log.warning("Empty file name when importing subscribe/publish stuff. Maybe the cancel button was pressed?")
                return
        else:
            file_path_name = filedialog.asksaveasfilename(initialdir=self.config_handler.get_last_used_directory(),
                                                          title="Export connection configuration",
                                                          defaultextension="json",
                                                          initialfile="{}-topics-content".format(self.connection_selector.get().replace(" ", "_")))
            if file_path_name == "":
                self.log.warning("Empty file name when exporting subscribe/publish stuff. Maybe the cancel button was pressed?")
                return

        self.config_handler.save_last_used_directory(file_path_name)
        self.file_input.delete(0, tk.END)
        self.file_input.insert(0, file_path_name)

        if not self.is_import:
            self.log.info("File selected for topic and message export: {}".format(file_path_name))
            self.on_checkbox()
            return

        if not os.path.isfile(file_path_name):
            self.log.error("The file picked for topic and message history import doesn't exist!", file_path_name)
            self.info_label["text"] = "Invalid file"
            self.info_label["foreground"] = "red"
            return

        try:
            with open(file_path_name, "r", encoding="utf-8") as configfile:
                history_dict = json.loads(configfile.read())
        except Exception as e:
            self.log.error("Failed to load topic and message history file", e, traceback.format_exc())

        self.update_items(history_dict)

        self.imported_history = history_dict
        self.log.info("File selected for communication config import: {}".format(file_path_name))

    def on_destroy(self, *args, **kwargs):
        self.grab_release()
        self.destroy()

    def update_items(self, history_dict):
        self.subscribe_topic_checkbox["text"] = "{} Subscribe topic history items".format(
            0 if "subscriptions" not in history_dict else len(history_dict["subscriptions"]))

        self.publish_topic_checkbox["text"] = "{} Publish topic history items".format(
            0 if "publish_topics" not in history_dict else len(history_dict["publish_topics"])
        )

        self.message_template_checkbox["text"] = "{} Message template items".format(
            0 if "stored_publishes" not in history_dict else len(history_dict["stored_publishes"]))

    def ok(self, *args, **kwargs):
        if self.is_import:
            connection_to_import_to = self.connection_selector.get()
            self.log.info("Attempting to import publish/subscribe history into {}".format(connection_to_import_to))
            config_to_update = self.config_handler.get_connection_config_dict(connection_to_import_to)
            try:
                for key in ("subscriptions", "stored_publishes"):
                    config_to_update[key].update(self.imported_history.get(key, {}))
                if "publish_topics" in self.imported_history:
                    for topic in self.imported_history["publish_topics"]:
                        if topic not in config_to_update["publish_topics"]:
                            config_to_update["publish_topics"].append(topic)
                self.config_handler.save_connection_dict(connection_to_import_to, config_to_update)
            except Exception as e:
                self.log.error("Failed to import subscribe/publish history", e, traceback.format_exc())
                messagebox.showerror("Error", "Error importing subscribe/publish history. See log for details")
            else:
                messagebox.showinfo("Success", "Publish/subscribe history imported successfully")
            self.on_destroy()
        else:
            connection_to_export = self.connection_selector.get()
            connection_config = self.config_handler.get_connection_config_dict(connection_to_export)
            self.log.info("Exporting subscription and publish history from {} into {}".format(connection_to_export,
                                                                                              self.file_input.get()))
            export_dict = {}
            if bool(self.subscribe_topic_selection.get()):
                export_dict["subscriptions"] = connection_config.get("subscriptions", {})
            if bool(self.publish_topic_selection.get()):
                export_dict["publish_topics"] = connection_config.get("publish_topics", {})
            if bool(self.message_template_selection.get()):
                export_dict["stored_publishes"] = connection_config.get("stored_publishes", {})

            try:
                with open(self.file_input.get(), "w", encoding="utf-8") as export_file:
                    export_file.write(json.dumps(export_dict, indent=2, ensure_ascii=False))
            except Exception as e:
                self.log.error("Failed to export publish and subscription history", e, traceback.format_exc())
                messagebox.showerror("Error exporting publish and subscription history", "See log for details")
            else:
                self.log.info("Successfully exported publish and subscription history")
                messagebox.showinfo("Success", "Subscription/publish history exported successfully")
            self.on_destroy()



"""
MQTTk - Lightweight graphical MQTT client and message analyser

Copyright (C) 2022  Máté Szabó

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

import sys
import os
import traceback
from pathlib import Path
import json
from tkinter import messagebox
from tkinter import filedialog
import mqttk.mqtt_fx_config_parser as configparser
from datetime import datetime

LOAD = "load"
SAVE = "save"

DEFAULT_CONFIGURATION = {
    "connections": {
        "test.mosquitto.org": {
            "connection_parameters": {
            "broker_addr": "test.mosquitto.org",
            "broker_port": "1883",
            "client_id": "272fb6890d1c4eefac4f46b39deb83fb",
            "user": "",
            "pass": "",
            "timeout": "10",
            "keepalive": "60",
            "mqtt_version": "3.1.1",
            "ssl": "Disabled",
            "ca_file": "",
            "cl_cert": "",
            "cl_key": ""
            },
        "subscriptions": {},
        "publish_topics": [],
        "stored_publishes": {},
        "last_subscribe_used": "#"
        }
    }
}


class ConfigHandler:
    def __init__(self, logger):
        """
        Config handler.

        configuration_dict = {
            "connections": {}
            "last_used_connection": connection name,
            "window_geometry: last used window geometry string,
            "autoscroll: true/false,
            "last_used_decoder": last used message decoder,
            "last_used_directory": last used directory for browsing files
        }

        configuration_dict[connections] = {
            "connection_profile_name": {
                "connection_parameters": {
                    "connection_parameter: value
                },
                "subscriptions": { list of previous subscriptions
                    "subscription_topic": {
                        "colour": color
                    }
                }
                "publish_topics": [] list of previous publishes
                "stored_publishes": {
                    "name": {
                        "topic": topic,
                        "qos": qos,
                        "payload": payload,
                        "retained": retained
                },
                "last_publish_used": last publish,
                "last_subscribe_used: last subscribe
            }
        }

        connection parameters:
        - broker_addr
        - broker_port
        - client_id
        - user
        - pass
        - timeout
        - keepalive
        - mqtt_version
        - ssl
        - ca_file
        - cl_cert
        - cl_key

        """
        self.configuration_dict = {}
        self.log_file = None
        self.wont_save = False
        self.first_start = True
        self.log = logger
        self.mqttfx_config_location = None
        self.config_file_manager(LOAD)

    def config_file_manager(self, action):
        if self.wont_save:
            return
        if sys.platform.startswith("win"):
            if self.first_start:
                self.log.info("Windoze platform detected")
                appdata_dir = os.getenv('LOCALAPPDATA')
                mqttfx_config_file = os.path.join(appdata_dir, "MQTT-FX", "mqttfx-config.xml")
                if os.path.isfile(mqttfx_config_file):
                    self.mqttfx_config_location = mqttfx_config_file
                    self.log.info("Found MQTT.fx configuration file", self.mqttfx_config_location)

            appdata_dir = os.getenv('LOCALAPPDATA')
            config_dir = os.path.join(appdata_dir, "MQTTk")
            config_file = os.path.join(appdata_dir, "MQTTk", "MQTTk-config.json")
            self.log_file = os.path.join(appdata_dir, "MQTTk", "MQTTk-log.txt")

        elif sys.platform.startswith("linux"):
            if self.first_start:
                self.log.info("Linux platform detected")
                home_dir = str(Path.home())
                mqttfx_config_file = os.path.join(home_dir, "MQTT-FX", "mqttfx-config.xml")
                if os.path.isfile(mqttfx_config_file):
                    self.mqttfx_config_location = mqttfx_config_file
                    self.log.info("Found MQTT.fx configuration file", self.mqttfx_config_location)

            home_dir = str(Path.home())
            config_dir = os.path.join(home_dir, ".config", "MQTTk")
            config_file = os.path.join(home_dir, ".config", "MQTTk", "MQTTk-config.json")
            self.log_file = os.path.join(home_dir, ".config", "MQTTk", "MQTTk-log.txt")

        elif sys.platform.startswith("darwin"):
            if self.first_start:
                self.log.info("MacOS platform detected")
                home_dir = str(Path.home())
                mqttfx_config_file = os.path.join(home_dir, "Library", "ApplicationSupport", "MQTT-FX", "mqttfx-config.xml")
                if os.path.isfile(mqttfx_config_file):
                    self.mqttfx_config_location = mqttfx_config_file
                    self.log.info("Found MQTT.fx configuration file", self.mqttfx_config_location)

            home_dir = str(Path.home())
            config_dir = os.path.join(home_dir, "Library", "ApplicationSupport", "MQTTk")
            config_file = os.path.join(home_dir, "Library", "ApplicationSupport", "MQTTk", "MQTTk-config.json")
            self.log_file = os.path.join(home_dir, "Library", "ApplicationSupport", "MQTTk", "MQTTk-log.txt")
        else:
            self.log.warning("Unsupported platform detected. Configuration file won't be saved! Use this thing at your own risk :(")
            self.wont_save = True

        self.first_start = False

        if self.wont_save:
            return

        if not os.path.isfile(config_file):
            self.configuration_dict = DEFAULT_CONFIGURATION
            self.first_start = True
            if not os.path.isdir(config_dir):
                os.makedirs(config_dir)
            with open(config_file, "w", encoding="utf-8") as configfile:
                configfile.write(json.dumps(self.configuration_dict, indent=2, ensure_ascii=False))

        else:
            if action == LOAD:
                with open(config_file, "r", encoding="utf-8") as configfile:
                    configuration = configfile.read()
                try:
                    self.configuration_dict = json.loads(configuration)
                except Exception as e:
                    self.log.error("Failed to load config", e)
            else:
                with open(config_file, "w", encoding="utf-8") as config_file:
                    try:
                        config_string = json.dumps(self.configuration_dict, indent=2, ensure_ascii=False)
                    except Exception as e:
                        self.log.error("Failed to save configuration", e)
                    else:
                        config_file.write(config_string)

    def get_connection_profiles(self):
        return list(self.configuration_dict.get("connections", {}).keys())

    def get_connection_config_dict(self, connection):
        return self.configuration_dict.get("connections", {}).get(connection, {})

    def remove_connection_config(self, connection_name):
        self.configuration_dict.get("connections", {}).pop(connection_name, None)
        self.config_file_manager(SAVE)

    def get_connection_broker_parameters(self, connection):
        return self.configuration_dict["connections"].get(connection, {}).get("connection_parameters", {})

    def save_connection_config(self, connection_name, connection_config):
        if "connections" not in self.configuration_dict:
            self.configuration_dict["connections"] = {}
        if connection_name not in self.configuration_dict["connections"]:
            self.configuration_dict["connections"][connection_name] = {
                "connection_parameters": {},
                "subscriptions": {},
                "publish_topics": [],
                "stored_publishes": {}
            }
        self.configuration_dict["connections"][connection_name]["connection_parameters"] = connection_config
        self.config_file_manager(SAVE)

    def save_connection_dict(self, connection_name, connection_config):
        if "connections" not in self.configuration_dict:
            self.configuration_dict["connections"] = {}
        self.configuration_dict["connections"][connection_name] = connection_config
        self.config_file_manager(SAVE)

    def add_subscription_history(self, connection, topic, colour):
        self.configuration_dict["connections"][connection]["subscriptions"][topic] = {
                "colour": colour
            }
        self.configuration_dict["connections"][connection]["last_subscribe_used"] = topic
        self.config_file_manager(SAVE)

    def get_subscription_history_list(self, connection):
        try:
            return list(self.configuration_dict.get("connections", {}).get(connection, {}).get("subscriptions", {}).keys())
        except AttributeError:
            try:
                self.configuration_dict["connections"][connection]["subscriptions"] = {}
            except Exception:
                self.log.error("Fatal subscription history incompatibility in the config")
            return None

    def get_subscription_colour(self, connection, topic):
        return self.configuration_dict.get("connections", {}).get(
            connection, {}).get("subscriptions", {}).get(topic, {}).get("colour", None)

    def get_window_geometry(self):
        return self.configuration_dict.get("window_geometry", None)

    def save_window_geometry(self, window_geometry):
        self.configuration_dict["window_geometry"] = window_geometry
        self.config_file_manager(SAVE)

    def get_last_used_connection(self):
        return self.configuration_dict.get("last_used_connection", "")

    def update_last_used_connection(self, connection):
        self.configuration_dict["last_used_connection"] = connection

    def get_autoscroll(self):
        return bool(self.configuration_dict.get("autoscroll", False))

    def save_autoscroll(self, value):
        self.configuration_dict["autoscroll"] = bool(value)
        self.config_file_manager(SAVE)

    def get_decompress(self):
        return bool(self.configuration_dict.get("decompress", False))

    def save_decompress(self, value):
        self.configuration_dict["decompress"] = bool(value)
        self.config_file_manager(SAVE)

    def get_decoder(self):
        return self.configuration_dict.get("decoder", "Plain data")

    def save_decoder(self, value):
        self.configuration_dict["decoder"] = value
        self.config_file_manager(SAVE)

    def delete_publish_history_item(self, connection, name):
        try:
            self.configuration_dict["connections"][connection]["stored_publishes"].pop(name, None)
        except Exception as e:
            self.log.warning("Failed to remove history publish item", e, connection, name)

    def get_publish_history(self, connection):
        return self.configuration_dict["connections"].get(connection, {}).get("stored_publishes", {})

    def save_publish_history_item(self, connection, name, config):
        try:
            if "stored_publishes" not in self.configuration_dict["connections"][connection]:
                self.configuration_dict["connections"][connection]["stored_publishes"] = {
                    name: config
                }
            else:
                self.configuration_dict["connections"][connection]["stored_publishes"][name] = config
            self.config_file_manager(SAVE)
        except Exception as e:
            self.log.warning("Exception saving publish history config", e)

    def get_publish_topic_history(self, connection):
        return self.configuration_dict["connections"].get(connection, {}).get("publish_topics", [])

    def save_publish_topic_history_item(self, connection, topic):
        try:
            new = True
            if "publish_topics" not in self.configuration_dict["connections"][connection]:
                self.configuration_dict["connections"][connection]["publish_topics"] = [topic]
            else:
                if topic not in self.configuration_dict["connections"][connection]["publish_topics"]:
                    self.configuration_dict["connections"][connection]["publish_topics"].append(topic)
                else:
                    new = False
            self.configuration_dict["connections"][connection]["last_publish_used"] = topic
            self.config_file_manager(SAVE)
            return new
        except Exception as e:
            self.log.error("Error saving publish topic history item", e)
        self.config_file_manager(SAVE)

    def get_last_publish_topic(self, connection):
        return self.configuration_dict.get("connections", {}).get(connection, {}).get("last_publish_used", "")

    def get_last_subscribe_used(self, connection):
        return self.configuration_dict.get("connections", {}).get(connection, {}).get("last_subscribe_used", "")

    def get_last_used_decoder(self):
        return self.configuration_dict.get("last_used_decoder", "None")

    def save_last_used_decoder(self, decoder):
        self.configuration_dict["last_used_decoder"] = decoder
        self.config_file_manager(SAVE)

    def import_mqttfx_config(self):
        if configparser.XMLTODICT is False:
            messagebox.showerror("Error", "Failed to import the xmltodict library. Please ensure all dependencies are installed!")
        if self.mqttfx_config_location is None:
            response = messagebox.askquestion("MQTT.fx config not found", "Couldn't find MQTT.fx configuration file. Would you like to browse the file?")
            if response == "no":
                return False
            mqtt_fx_config_file = filedialog.askopenfilename(initialdir=str(Path.home()),
                                                             title="Select MQTT.fx configuration file")
            if not os.path.isfile(mqtt_fx_config_file):
                messagebox.showerror("Error", "File cannot be found")
                return False
        else:
            mqtt_fx_config_file = self.mqttfx_config_location

        error, response = configparser.parse_mqttfx_xml(mqtt_fx_config_file)
        if error:
            messagebox.showerror("Error", response)
            return False

        error, response = configparser.parse_mqttfx_config(response, self.configuration_dict)
        if error:
            messagebox.showerror("Error", "Failed to parse MQTT.fx config dict: {}".format(response))
            return False

        messagebox.showinfo("Success!", "Successfully imported MQTT.fx configuration! Please double check SSL configuration!")
        self.config_file_manager(SAVE)
        self.configuration_dict = response
        return True

    def get_last_used_directory(self):
        if "last_used_directory" not in self.configuration_dict:
            self.configuration_dict["last_used_directory"] = Path.home()
        if not os.path.isdir(self.configuration_dict["last_used_directory"]):
            self.configuration_dict["last_used_directory"] = Path.home()
        return self.configuration_dict["last_used_directory"]

    def save_last_used_directory(self, directory):
        head, tail = os.path.split(directory)
        self.configuration_dict["last_used_directory"] = head
        self.config_file_manager(SAVE)

    def add_log_message(self, message):
        if self.log_file is not None:
            if not os.path.isfile(self.log_file):
                with open(self.log_file, 'w', encoding="utf-8") as logfile:
                    logfile.write("New log file started {}{}".format(datetime.now().strftime("%Y/%m/%d, %H:%M:%S.%f"),
                                                                     os.linesep))
            with open(self.log_file, 'a', encoding="utf-8") as logfile:
                logfile.write(message)

    def save_export_encode_selection(self, value):
        self.configuration_dict["export_encoding"] = value
        self.config_file_manager(SAVE)

    def get_export_encode_selection(self):
        return self.configuration_dict.get("export_encoding", 1)

    def get_resubscribe(self, connection):
        return self.configuration_dict.get("connections", {}).get(connection, {}).get("connection_parameters", {}).get("resubscribe", 0)

    def get_resubscribe_topics(self, connection):
        return self.configuration_dict.get("connections", {}).get(connection, {}).get("resubscribe_topics", [])

    def save_resubscribe_topics(self, connection, topics):
        try:
            self.configuration_dict["connections"][connection]["resubscribe_topics"] = topics
            self.config_file_manager(SAVE)
        except Exception as e:
            self.log.error("Failed to save resubscribe topics:", e)



"""
MQTTk - Lightweight graphical MQTT client and message analyser

Copyright (C) 2022  Máté Szabó

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
import tkinter as tk
import tkinter.ttk as ttk
import traceback
from tkinter import filedialog
from mqttk.widgets.scroll_frame import ScrollFrame
from mqttk.constants import SSL_LIST, MQTT_VERSION_LIST
import uuid
from functools import partial
from mqttk.helpers import validate_name, validate_int, get_clear_combobox_selection_function, clear_combobox_selection


class ConnectionFrame(ttk.Frame):
    def __init__(self, container, connection_name, on_select_callback=None, *args, **kwargs):
        super().__init__(container, *args, **kwargs)
        self.container = container
        self.connection_name = connection_name
        self.on_select_callback = on_select_callback
        self["relief"] = "groove"
        self["borderwidth"] = 2
        self.connection = ttk.Label(self)
        self.connection["text"] = connection_name
        self.connection.pack(fill=tk.X, expand=1)
        self.bind("<Button-1>", self.on_click)
        self.connection.bind("<Button-1>", self.on_click)

    def on_click(self, *args, **kwargs):
        if self.on_select_callback is not None:
            self.configure(style="Selected.TFrame")
            self.connection.configure(style="Selected.TLabel")
            self.on_select_callback(self.connection_name)

    def on_unselect(self):
        self.configure(style="TFrame")
        self.connection.configure(style="TLabel")
        self.update()


class ConfigurationWindow(tk.Toplevel):
    def __init__(self, master, config_handler, config_update_callback, logger, icon, selected_connection):
        super().__init__(master=master)
        self.transient(master)
        self.master = master
        self.config_handler = config_handler
        self.currently_selected_connection = None
        self.currently_selected_connection_dict = {}
        self.config_update_callback = config_update_callback
        self.log = logger
        self.just_opened = True

        self.grab_set()
        self.title("Connection configuration")
        width = 1000
        height = 600
        screenwidth = self.winfo_screenwidth()
        screenheight = self.winfo_screenheight()
        alignstr = '%dx%d+%d+%d' % (width, height, (screenwidth - width) / 2, (screenheight - height) / 2)
        self.geometry(alignstr)

        self.protocol("WM_DELETE_WINDOW", self.on_destroy)
        self.bind("<Escape>", self.cancel)
        vcmd = (self.register(validate_int),
                '%d', '%i', '%P', '%s', '%S', '%v', '%V', '%W')

        self.iconphoto(False, icon)

        self.background_frame = ttk.Frame(self)
        self.background_frame.pack(fill='both', expand=1)

        # Connections frame
        self.connections_frame = ttk.Frame(self.background_frame, relief="sunken")
        self.connections_frame.pack(side=tk.LEFT, anchor="w", fill="y", padx=3, pady=3)
        self.connections_listbox = ScrollFrame(self.connections_frame)
        self.connections_listbox.pack(fill='both', padx=3, pady=3, expand=1)
        self.add_connection_button = ttk.Button(self.connections_frame, text="Add", command=self.new_connection)
        self.add_connection_button.pack(padx=3, pady=3, side="right")
        self.remove_connection_button = ttk.Button(self.connections_frame, text="Remove", command=self.on_remove)
        self.remove_connection_button.pack(padx=3, pady=3, side="right")

        # Connection configuration frame
        self.connection_configuration_frame = ttk.Frame(self.background_frame, relief="groove", borderwidth=2)
        self.connection_configuration_frame.pack(side=tk.RIGHT, fill="both", pady=3, padx=3, expand=1)

        # Profile name
        self.profile_name_frame = ttk.Frame(self.connection_configuration_frame)
        self.profile_name_frame.pack(fill="x")
        self.profile_label = ttk.Label(self.profile_name_frame, width=15, anchor="e", text="Profile name")
        self.profile_label.pack(side=tk.LEFT, anchor="w", padx=2, pady=4)
        self.profile_name_input = ttk.Entry(self.profile_name_frame)
        self.profile_name_input.pack(side=tk.LEFT, padx=2)

        # Broker address
        self.broker_address_frame = ttk.Frame(self.connection_configuration_frame)
        self.broker_address_frame.pack(fill="x")
        self.broker_label = ttk.Label(self.broker_address_frame, width=15, anchor="e", text="Broker address")
        self.broker_label.pack(side=tk.LEFT, anchor="w", padx=2, pady=4)
        self.broker_address_input = ttk.Entry(self.broker_address_frame, background="white")
        self.broker_address_input.pack(side=tk.LEFT, padx=2)

        # Broker port
        self.broker_port_frame = ttk.Frame(self.connection_configuration_frame)
        self.broker_port_frame.pack(fill="x")
        self.broker_port_label = ttk.Label(self.broker_port_frame, width=15, anchor="e", text="Broker port")
        self.broker_port_label.pack(side=tk.LEFT, anchor="w", padx=2, pady=4)
        self.broker_port_name_input = ttk.Entry(self.broker_port_frame)
        self.broker_port_name_input.configure(validate="all", validatecommand=vcmd)
        self.broker_port_name_input.pack(side=tk.LEFT, padx=2)

        # Client ID
        self.client_id_frame = ttk.Frame(self.connection_configuration_frame)
        self.client_id_frame.pack(fill="x")
        self.client_id_label = ttk.Label(self.client_id_frame, width=15, anchor="e", text="Client ID")
        self.client_id_label.pack(side=tk.LEFT, anchor="w", padx=2, pady=4)
        self.client_id_input = ttk.Entry(self.client_id_frame)
        self.client_id_input.pack(side=tk.LEFT, padx=2)
        self.client_id_generate_button = ttk.Button(self.client_id_frame,
                                                    text="Generate client ID",
                                                    command=self.on_generate_client_id)
        self.client_id_autogen = tk.IntVar()
        self.client_id_generate_button.pack(side=tk.LEFT, padx=2, pady=2)
        self.client_id_autogen_checkbox = ttk.Checkbutton(self.client_id_frame,
                                                          text="Generate ID before connecting",
                                                          variable=self.client_id_autogen,
                                                          offvalue=0,
                                                          onvalue=1,
                                                          command=self.on_client_id_autogen)
        self.client_id_autogen_checkbox.pack(side=tk.LEFT, padx=2, pady=2)

        # username
        self.username_frame = ttk.Frame(self.connection_configuration_frame)
        self.username_frame.pack(fill="x")
        self.username_label = ttk.Label(self.username_frame, width=15, anchor="e", text="Username")
        self.username_label.pack(side=tk.LEFT, anchor="w", padx=2, pady=4)
        self.username_input = ttk.Entry(self.username_frame)
        self.username_input.pack(side=tk.LEFT, padx=2)

        # password
        self.password_frame = ttk.Frame(self.connection_configuration_frame)
        self.password_frame.pack(fill="x")
        self.password_label = ttk.Label(self.password_frame, width=15, anchor="e", text="Password")
        self.password_label.pack(side=tk.LEFT, anchor="w", padx=2, pady=4)
        self.password_input = ttk.Entry(self.password_frame, show='*')
        self.password_input.pack(side=tk.LEFT, padx=2)

        # timeout
        self.timeout_frame = ttk.Frame(self.connection_configuration_frame)
        self.timeout_frame.pack(fill="x")
        self.timeout_label = ttk.Label(self.timeout_frame, width=15, anchor="e", text="Timeout")
        self.timeout_label.pack(side=tk.LEFT, anchor="w", padx=2, pady=4)
        self.timeout_input = ttk.Entry(self.timeout_frame)
        self.timeout_input.configure(validate="all", validatecommand=vcmd)
        self.timeout_input.pack(side=tk.LEFT, padx=2)

        # keepalive
        self.keepalive_frame = ttk.Frame(self.connection_configuration_frame)
        self.keepalive_frame.pack(fill="x")
        self.keepalive_label = ttk.Label(self.keepalive_frame, width=15, anchor="e", text="Keepalive")
        self.keepalive_label.pack(side=tk.LEFT, anchor="w", padx=2, pady=4)
        self.keepalive_input = ttk.Entry(self.keepalive_frame)
        self.keepalive_input.configure(validate="all", validatecommand=vcmd)
        self.keepalive_input.pack(side=tk.LEFT, padx=2)

        # MQTT version
        self.version_frame = ttk.Frame(self.connection_configuration_frame)
        self.version_frame.pack(fill="x")
        self.version_label = ttk.Label(self.version_frame, width=15, anchor="e", text="MQTT version")
        self.version_label.pack(side=tk.LEFT, anchor="w", padx=2, pady=4)
        self.version_input = ttk.Combobox(self.version_frame, exportselection=False, values=MQTT_VERSION_LIST)
        self.version_input.bind("<<ComboboxSelected>>", get_clear_combobox_selection_function(self.version_input))
        self.version_input.pack(side=tk.LEFT, padx=2)
        self.version_input.current(1)
        self.version_input["state"] = "readonly"

        # SSL
        self.ssl_state_frame = ttk.Frame(self.connection_configuration_frame)
        self.ssl_state_frame.pack(fill="x")
        self.ssl_state_label = ttk.Label(self.ssl_state_frame, width=15, anchor="e", text="SSL")
        self.ssl_state_label.pack(side=tk.LEFT, anchor="w", padx=2, pady=4)
        self.ssl_state_input = ttk.Combobox(self.ssl_state_frame,
                                            exportselection=False,
                                            values=SSL_LIST,
                                            state="readonly")
        self.ssl_state_input.bind("<<ComboboxSelected>>", get_clear_combobox_selection_function(self.ssl_state_input))
        self.ssl_state_input.pack(side=tk.LEFT, padx=2)

        # ca file
        self.ca_file_frame = ttk.Frame(self.connection_configuration_frame)
        self.ca_file_frame.pack(fill="x")
        self.ca_file_label = ttk.Label(self.ca_file_frame, width=15, anchor="e", text="CA file")
        self.ca_file_label.pack(side=tk.LEFT, anchor="w", padx=2, pady=4)
        self.ca_file_input = ttk.Entry(self.ca_file_frame, width=40)
        self.ca_file_input.pack(side=tk.LEFT, padx=2)
        self.ca_browser_button = ttk.Button(self.ca_file_frame,
                                            width=3,
                                            text="...",
                                            command=partial(self.browse_file, target_entry=self.ca_file_input))
        self.ca_browser_button.pack(side=tk.LEFT, padx=2, pady=2)

        # client cert
        self.cl_cert_file_frame = ttk.Frame(self.connection_configuration_frame)
        self.cl_cert_file_frame.pack(fill="x")
        self.cl_cert_file_label = ttk.Label(self.cl_cert_file_frame,
                                            width=15,
                                            anchor="e",
                                            text="Client certificate file")
        self.cl_cert_file_label.pack(side=tk.LEFT, anchor="w", padx=2, pady=4)
        self.cl_cert_file_input = ttk.Entry(self.cl_cert_file_frame, width=40)
        self.cl_cert_file_input.pack(side=tk.LEFT, padx=2)
        self.cl_cert_browser_button = ttk.Button(self.cl_cert_file_frame,
                                                 width=3,
                                                 text="...",
                                                 command=partial(self.browse_file,
                                                                 target_entry=self.cl_cert_file_input))
        self.cl_cert_browser_button.pack(side=tk.LEFT, padx=2, pady=2)

        # client_key
        self.cl_key_file_frame = ttk.Frame(self.connection_configuration_frame)
        self.cl_key_file_frame.pack(fill="x")
        self.cl_key_file_label = ttk.Label(self.cl_key_file_frame, width=15, anchor="e", text="Client key file")
        self.cl_key_file_label.pack(side=tk.LEFT, anchor="w", padx=2, pady=4)
        self.cl_key_file_input = ttk.Entry(self.cl_key_file_frame, width=40)
        self.cl_key_file_input.pack(side=tk.LEFT, padx=2)
        self.cl_key_browser_button = ttk.Button(self.cl_key_file_frame,
                                                width=3,
                                                text="...",
                                                command=partial(self.browse_file, target_entry=self.cl_key_file_input))
        self.cl_key_browser_button.pack(side=tk.LEFT, padx=2, pady=2)

        self.ssl_state_input.bind("<<ComboboxSelected>>", self.ssl_state_change)
        self.ssl_state_input.current(0)
        self.ssl_state_change(None)

        # Client ID
        self.resubscribe_frame = ttk.Frame(self.connection_configuration_frame)
        self.resubscribe_frame.pack(fill="x", pady=4)
        self.divider_label = ttk.Label(self.resubscribe_frame, width=15, anchor="e",
                                       text="Preferences")
        self.divider_label.pack(side=tk.LEFT, padx=2, pady=4)
        self.resubscribe_state = tk.IntVar()
        self.resubscribe_checkbox = ttk.Checkbutton(self.resubscribe_frame,
                                                    text="Automatically re-suscribe to the last used topics after connection",
                                                    variable=self.resubscribe_state,
                                                    offvalue=0,
                                                    onvalue=1,
                                                    command=self.on_client_id_autogen)
        self.resubscribe_checkbox.pack(side=tk.LEFT, padx=2, pady=2)

        self.button_frame = ttk.Frame(self.connection_configuration_frame)
        self.button_frame.pack(side='bottom', fill='x', anchor='s', expand=1)
        self.ok_button = ttk.Button(self.button_frame, text="OK", command=self.ok)
        self.ok_button.pack(side=tk.RIGHT, pady=5, padx=5)
        self.apply_button = ttk.Button(self.button_frame, text="Apply", command=self.apply)
        self.apply_button.pack(side=tk.RIGHT, pady=5, padx=5)
        self.cancel_button = ttk.Button(self.button_frame, text="Cancel", command=self.cancel)
        self.cancel_button.pack(side=tk.RIGHT, pady=5, padx=5)

        self.all_config_state_change("disabled")

        self.profiles_widgets = {}

        connection_profiles = self.config_handler.get_connection_profiles()
        for connection_profile in sorted(connection_profiles):
            self.add_profile_widget(connection_profile)
            if connection_profile == selected_connection:
                self.profiles_widgets[connection_profile].on_click()
        self.focus_set()
        self.wait_window(self)

    def on_generate_client_id(self):
        self.client_id_input.delete(0, tk.END)
        self.client_id_input.insert(0, str(uuid.uuid4()).replace("-", ""))

    def on_client_id_autogen(self, *args, **kwargs):
        self.client_id_input.configure(state="normal" if self.client_id_autogen.get() == 0 else "disabled")
        self.client_id_generate_button.configure(state="normal" if self.client_id_autogen.get() == 0 else "disabled")

    def apply(self, *args, **kwargs):
        self.save_current_config()

    def ok(self, *args, **kwargs):
        self.save_current_config()
        self.on_destroy()

    def cancel(self, *args, **kwargs):
        self.on_destroy()

    def on_remove(self):
        if self.currently_selected_connection is not None:
            self.config_handler.remove_connection_config(self.currently_selected_connection)
            self.profiles_widgets[self.currently_selected_connection].pack_forget()
            self.profiles_widgets.pop(self.currently_selected_connection)
            if len(self.profiles_widgets) != 0:
                self.connection_selected(list(self.profiles_widgets.keys())[0])

    def save_current_config(self):
        if self.currently_selected_connection is None:
            return
        config_dict = {
            "broker_addr": self.broker_address_input.get(),
            "broker_port": self.broker_port_name_input.get(),
            "client_id": self.client_id_input.get(),
            "client_id_autogen": self.client_id_autogen.get(),
            "user": self.username_input.get(),
            "pass": self.password_input.get(),
            "timeout": self.timeout_input.get(),
            "keepalive": self.keepalive_input.get(),
            "mqtt_version": self.version_input.get(),
            "ssl": self.ssl_state_input.get(),
            "ca_file": self.ca_file_input.get(),
            "cl_cert": self.cl_cert_file_input.get(),
            "cl_key": self.cl_key_file_input.get(),
            "resubscribe": self.resubscribe_state.get(),
        }
        self.currently_selected_connection_dict = config_dict
        self.config_handler.save_connection_config(self.profile_name_input.get(), config_dict)

        # If there was a name change, delete the old from the config file via the config handler and update the widget
        if self.profile_name_input.get() != self.currently_selected_connection:
            self.config_handler.remove_connection_config(self.currently_selected_connection)
            self.profiles_widgets[self.currently_selected_connection].connection_name = self.profile_name_input.get()
            self.profiles_widgets[self.currently_selected_connection].connection["text"] = self.profile_name_input.get()
            self.profiles_widgets[self.profile_name_input.get()] = self.profiles_widgets[
                self.currently_selected_connection]
            self.config_handler.remove_connection_config(self.currently_selected_connection)
            self.connection_selected(self.currently_selected_connection)

        self.config_handler.update_last_used_connection(self.currently_selected_connection)

    def browse_file(self, target_entry):
        file_path_name = filedialog.askopenfilename(initialdir=self.config_handler.get_last_used_directory(),
                                                    title="Select CA file")
        if file_path_name == "":
            return
        self.config_handler.save_last_used_directory(file_path_name)
        target_entry.delete(0, tk.END)
        target_entry.insert(0, file_path_name)

    def add_profile_widget(self, connection_profile):
        self.profiles_widgets[connection_profile] = ConnectionFrame(self.connections_listbox.viewPort,
                                                                    connection_profile,
                                                                    self.connection_selected)
        self.profiles_widgets[connection_profile].pack(fill=tk.X, expand=1, padx=2, pady=2)

    def new_connection(self):
        name = validate_name("New connection", self.profiles_widgets)
        self.add_profile_widget(name)
        self.profiles_widgets[name].on_click(None)
        self.connection_selected(name)
        self.currently_selected_connection_dict = {}

    def connection_selected(self, connection_name):
        if self.currently_selected_connection != connection_name:
            try:
                self.profiles_widgets[self.currently_selected_connection].on_unselect()
            except Exception as e:
                if not self.just_opened:
                    self.log.warning("Exception deselecting profile widget, maybe there wasn't one selected?", e,
                                 self.currently_selected_connection, connection_name)
                    self.just_opened = False
            try:
                self.all_config_state_change("normal")
                self.currently_selected_connection_dict = self.config_handler.get_connection_config_dict(
                    connection_name).get("connection_parameters", {})
                self.profile_name_input.delete(0, tk.END)
                self.profile_name_input.insert(0, connection_name)
                self.broker_address_input.delete(0, tk.END)
                self.broker_address_input.insert(0, self.currently_selected_connection_dict.get("broker_addr",
                                                                                                "mqtt.example.com"))
                self.broker_port_name_input.delete(0, tk.END)
                self.broker_port_name_input.insert(0, self.currently_selected_connection_dict.get("broker_port",
                                                                                                  "1883"))
                self.client_id_input.delete(0, tk.END)
                self.client_id_input.insert(0, self.currently_selected_connection_dict.get("client_id", "MQTTk_Client"))
                self.client_id_autogen.set(self.currently_selected_connection_dict.get("client_id_autogen", 0))
                self.username_input.delete(0, tk.END)
                self.username_input.insert(0, self.currently_selected_connection_dict.get("user", ""))
                self.password_input.delete(0, tk.END)
                self.password_input.insert(0, self.currently_selected_connection_dict.get("pass", ""))
                self.timeout_input.delete(0, tk.END)
                self.timeout_input.insert(0, self.currently_selected_connection_dict.get("timeout", "10"))
                self.keepalive_input.delete(0, tk.END)
                self.keepalive_input.insert(0, self.currently_selected_connection_dict.get("keepalive", "60"))

                mqtt_version = self.currently_selected_connection_dict.get("mqtt_version", "")
                if mqtt_version in MQTT_VERSION_LIST:
                    self.version_input.current(MQTT_VERSION_LIST.index(mqtt_version))
                else:
                    self.version_input.current(1)

                self.ssl_state_input.current(SSL_LIST.index("Self-signed certificate"))
                self.ssl_state_change(None)
                ssl = self.currently_selected_connection_dict.get("ssl", "")
                if ssl in SSL_LIST:
                    self.ssl_state_input.current(SSL_LIST.index(ssl))
                else:
                    self.ssl_state_input.current(0)

                self.ca_file_input.delete(0, tk.END)
                self.ca_file_input.insert(0, self.currently_selected_connection_dict.get("ca_file", ""))
                self.cl_cert_file_input.delete(0, tk.END)
                self.cl_cert_file_input.insert(0, self.currently_selected_connection_dict.get("cl_cert", ""))
                self.cl_key_file_input.delete(0, tk.END)
                self.cl_key_file_input.insert(0, self.currently_selected_connection_dict.get("cl_key", ""))
                self.ssl_state_change(None)
                self.resubscribe_state.set(self.currently_selected_connection_dict.get("resubscribe", 0))
            except Exception as e:
                self.all_config_state_change("disabled")
                self.log.exception("Failed to load connection!", e, traceback.print_exc())
            else:
                self.currently_selected_connection = connection_name
                self.all_config_state_change("normal")
                self.on_client_id_autogen()

    def all_config_state_change(self, state):
        self.profile_name_input.configure(state=state)
        self.broker_address_input.configure(state=state)
        self.broker_port_name_input.configure(state=state)
        self.client_id_input.configure(state=state)
        self.client_id_generate_button.configure(state=state)
        self.client_id_autogen_checkbox.configure(state=state)
        self.username_input.configure(state=state)
        self.password_input.configure(state=state)
        self.timeout_input.configure(state=state)
        self.keepalive_input.configure(state=state)

    def cert_state_change(self, ca, clc, clk):
        self.ca_file_input.configure(state=ca)
        self.ca_browser_button.configure(state=ca)
        self.cl_cert_file_input.configure(state=clc)
        self.cl_cert_browser_button.configure(state=clc)
        self.cl_key_file_input.configure(state=clk)
        self.cl_key_browser_button.configure(state=clk)

    def ssl_state_change(self, _):
        clear_combobox_selection(combobox_instance=self.ssl_state_input)
        if self.ssl_state_input.get() == "Disabled" or self.ssl_state_input.get() == "CA signed server certificate":
            self.cert_state_change("disabled", "disabled", "disabled")
        elif self.ssl_state_input.get() == "CA certificate file":
            self.cert_state_change("normal", "disabled", "disabled")
        else:
            self.cert_state_change("normal", "normal", "normal")

    def on_destroy(self, *args, **kwargs):
        self.grab_release()
        if self.config_update_callback is not None:
            self.config_update_callback()
        self.destroy()



"""
MQTTk - Lightweight graphical MQTT client and message analyser

Copyright (C) 2022  Máté Szabó

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
import os
import tkinter as tk
import tkinter.ttk as ttk
import time
import sys
import traceback
from functools import partial
from datetime import datetime

from mqttk.constants import CONNECT, COLOURS


class TopicBrowser(ttk.Frame):
    def __init__(self, master, config_handler, log, root, *args, **kwargs):
        super().__init__(master=master, *args, **kwargs)
        self.config_handler = config_handler
        self.log = log
        self.subscription_frames = {}
        self.color_carousel = -1
        self.current_connection = None
        self.last_connection = None
        self.current_subscription = None
        self.root = root

        self.mqtt_manager = None
        self.message_id_counter = 0
        self.individual_topics = 0

        # Subscribe frame
        self.topic_browser_bar_frame = ttk.Frame(self, height=1)
        self.topic_browser_bar_frame.pack(anchor="nw", side=tk.TOP, fill=tk.X)
        # Subscribe selector combobox
        self.subscribe_selector = ttk.Combobox(self.topic_browser_bar_frame, width=30, exportselection=False)
        self.subscribe_selector.pack(side=tk.LEFT, padx=3, pady=3)
        self.subscribe_selector["values"] = []
        # Subscribe button
        self.browse_button = ttk.Button(self.topic_browser_bar_frame, width=10, text="Browse")
        self.browse_button.pack(side=tk.LEFT, padx=3, pady=3)
        self.browse_button["command"] = self.add_subscription
        # Stop button
        self.stop_button = ttk.Button(self.topic_browser_bar_frame, width=10, text="Stop", state="disabled")
        self.stop_button.pack(side=tk.LEFT, padx=3, pady=3)
        self.stop_button["command"] = self.on_unsubscribe

        self.stat_label = ttk.Label(self.topic_browser_bar_frame)
        self.stat_label.pack(side=tk.LEFT, padx=3, pady=3)

        # Flush messages button
        self.flush_messages_button = ttk.Button(self.topic_browser_bar_frame, text="Clear topics")
        self.flush_messages_button.pack(side=tk.RIGHT, padx=3)
        self.flush_messages_button["command"] = self.flush_messages
        # Filter retained checkbox
        self.filter_retained = tk.IntVar()
        self.filter_retained_checkbox = ttk.Checkbutton(self.topic_browser_bar_frame,
                                                        text="Ignore retained messages",
                                                        variable=self.filter_retained,
                                                        offvalue=0,
                                                        onvalue=1)
        self.filter_retained_checkbox.pack(side=tk.RIGHT, padx=3)

        self.treeview_frame = ttk.Frame(self)
        self.treeview_frame.pack(expand=1, fill="both", pady=2, padx=2)
        self.topic_treeview = ttk.Treeview(self.treeview_frame, columns=("qos", "retained", "last_message", "payload"), show="tree headings")
        self.topic_treeview.heading('#0', text='Topic')
        self.topic_treeview.column('#0', minwidth=300, width=300, stretch=tk.NO)
        self.topic_treeview.heading('qos', text='QoS')
        self.topic_treeview.column('qos', minwidth=50, width=50, stretch=tk.NO)
        self.topic_treeview.heading('retained', text='Retained')
        self.topic_treeview.column('retained', minwidth=70, width=80, stretch=tk.NO)
        self.topic_treeview.heading('last_message', text='Last message')
        self.topic_treeview.column('last_message', minwidth=70, width=90, stretch=tk.NO)
        self.topic_treeview.heading('payload', text='Payload')
        self.topic_treeview.column('payload', minwidth=300, width=900, stretch=tk.NO)
        if sys.platform == "darwin":
            self.topic_treeview.bind("<Button-2>", self.popup)
        if sys.platform == "linux":
            self.topic_treeview.bind("<Button-3>", self.popup)
        if sys.platform == "win32":
            self.topic_treeview.bind("<Button-3>", self.popup)

        self.vertical_scrollbar = ttk.Scrollbar(self.treeview_frame, orient="vertical", command=self.topic_treeview.yview)
        self.vertical_scrollbar.pack(side=tk.RIGHT, fill="y")
        self.topic_treeview.configure(yscrollcommand=self.vertical_scrollbar.set)
        self.topic_treeview.pack(fill="both", side=tk.LEFT, expand=1)

        self.horizontal_scrollbar = ttk.Scrollbar(self, orient="horizontal", command=self.topic_treeview.xview)
        self.horizontal_scrollbar.pack(side=tk.BOTTOM, fill="x")
        self.topic_treeview.configure(xscrollcommand=self.horizontal_scrollbar.set)

        self.popup_menu = tk.Menu(self, tearoff=0)
        self.popup_menu.add_command(label="Copy topic", command=self.copy_topic)
        self.popup_menu.add_command(label="Copy payload", command=self.copy_payload)

    def interface_toggle(self, connection_state, mqtt_manager, current_connection):
        # Subscribe tab items
        self.mqtt_manager = mqtt_manager
        if connection_state != CONNECT:
            self.last_connection = self.current_connection
        else:
            if self.last_connection != current_connection:
                self.flush_messages()

        self.current_connection = current_connection
        self.browse_button.configure(state="normal" if connection_state is CONNECT else "disabled")
        self.subscribe_selector.configure(state="normal" if connection_state is CONNECT else "disabled")
        self.current_subscription = None

    def get_color(self, topic):
        colour = self.config_handler.get_subscription_colour(self.current_connection, topic)
        if colour is not None:
            return colour
        self.color_carousel += 1
        if self.color_carousel > len(COLOURS):
            self.color_carousel = 0
        return COLOURS[self.color_carousel]

    def add_subscription(self):
        topic = self.subscribe_selector.get()
        if topic != "":
            try:
                callback = partial(self.on_mqtt_message, subscription_pattern=topic)
                callback.__name__ = "MyCallback"  # This is to fix some weird behaviour of the paho client on linux
                self.mqtt_manager.add_subscription(topic_pattern=topic,
                                                   on_message_callback=callback)
            except Exception as e:
                self.log.exception("Failed to subscribe!", e)
                return
            if self.subscribe_selector["values"] == "":
                self.subscribe_selector["values"] = [topic]
            elif topic not in self.subscribe_selector['values']:
                self.subscribe_selector['values'] += (topic,)
            self.config_handler.add_subscription_history(self.current_connection,
                                                         topic,
                                                         self.get_color(topic))
        self.current_subscription = topic
        self.browse_button["state"] = "disabled"
        self.stop_button["state"] = "normal"

    def load_subscription_history(self):
        self.subscribe_selector.configure(
            values=self.config_handler.get_subscription_history_list(self.current_connection))
        self.subscribe_selector.set(self.config_handler.get_last_subscribe_used(self.current_connection))

    def on_mqtt_message(self, _, __, msg, subscription_pattern):
        try:
            if bool(self.filter_retained.get()) and msg.retain == 1:
                return

            try:
                payload_decoded = str(msg.payload.decode("utf-8"))
            except Exception:
                payload_decoded = msg.payload

            time_string = datetime.fromtimestamp(time.time()).strftime("%H:%M:%S")
            topic_split = msg.topic.split("/")

            # Fix anomaly when someone thinks MQTT is linux and starts the topic with /...
            if msg.topic.startswith("/"):
                topic_split[0] = "/"

            if 1 < len(topic_split):
                if topic_split[0] not in self.topic_treeview.get_children(""):
                    try:
                        self.topic_treeview.insert("", "end", topic_split[0], text=topic_split[0])
                    except Exception:
                        # print(msg.topic, topic_split[0], self.topic_treeview.get_children(""))
                        raise
                for i in range(1, len(topic_split)-1):
                    parent_topic = "/".join(topic_split[0:i])
                    topic = "/".join(topic_split[0:i+1])
                    if topic not in self.topic_treeview.get_children(parent_topic):
                        self.topic_treeview.insert(parent_topic, "end", topic, text=topic_split[i])

            parent_topic = "/".join(topic_split[0:-1])
            topic = "/".join(topic_split)
            if topic not in self.topic_treeview.get_children(parent_topic):
                self.update_individual_topics()
                self.topic_treeview.insert(parent_topic,
                                           "end",
                                           "/".join(topic_split),
                                           text=topic_split[-1],
                                           values=(msg.qos,
                                                   "RETAINED" if msg.retain == 1 else "",
                                                   time_string,
                                                   payload_decoded))
            else:
                self.topic_treeview.set(topic, "qos", msg.qos)
                self.topic_treeview.set(topic, "retained", "RETAINED" if msg.retain == 1 else "")
                self.topic_treeview.set(topic, "last_message", time_string)
                self.topic_treeview.set(topic, "payload", payload_decoded)

        except Exception as e:
            self.log.exception("Exception inserting new message to treeview",
                               os.linesep, msg.topic, msg.payload, os.linesep, e, traceback.format_exc())

    def on_unsubscribe(self):
        try:
            self.mqtt_manager.unsubscribe(self.current_subscription)
        except Exception as e:
            self.log.warning("Failed to unsubscribe", self.current_subscription, "maybe a failed subscription?")
        self.current_subscription = None
        self.stop_button["state"] = "disabled"
        self.browse_button["state"] = "normal"

    def flush_messages(self):
        self.update_individual_topics(0)
        for child in self.topic_treeview.get_children():
            self.topic_treeview.delete(child)

    def update_individual_topics(self, value=None):
        if value is not None:
            self.individual_topics = value
        else:
            self.individual_topics += 1
        self.stat_label["text"] = "{} individual topics mapped".format(self.individual_topics)

    def copy_topic(self, *args, **kwargs):
        try:
            topic = self.topic_treeview.selection()[0]
        except Exception:
            pass
        else:
            self.root.clipboard_clear()
            self.root.clipboard_append(topic)

    def copy_payload(self, *args, **kwargs):
        try:
            selection = self.topic_treeview.selection()[0]
            values = self.topic_treeview.item(selection).get("values", [])
            payload = values[3]
        except Exception:
            pass
        else:
            self.root.clipboard_clear()
            self.root.clipboard_append(payload)

    def popup(self, event, *args, **kwargs):
        try:
            self.popup_menu.tk_popup(event.x_root, event.y_root, 0)
        except Exception as e:
            pass
        finally:
            self.popup_menu.grab_release()



"""
MQTTk - Lightweight graphical MQTT client and message analyser

Copyright (C) 2022  Máté Szabó

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

import tkinter as tk
import tkinter.ttk as ttk
from mqttk.widgets.scrolled_text import CustomScrolledText


class LogTab(ttk.Frame):
    def __init__(self, master, log, *args, **kwargs):
        super().__init__(master=master, *args, **kwargs)
        self.master = master
        self.log = log

        self.log_options = ttk.Frame(self)
        self.allow_paho_debug = tk.IntVar()
        self.allow_paho_debug_button = ttk.Checkbutton(self.log_options,
                                                       text="Enable MQTT client debug messages",
                                                       variable=self.allow_paho_debug,
                                                       onvalue=1,
                                                       offvalue=0)
        self.allow_paho_debug_button.pack(side=tk.RIGHT, padx=4, pady=4)
        self.allow_paho_debug_button['command'] = self.on_paho_debug_toggle
        self.log_options.pack(fill="x")

        self.log_output = CustomScrolledText(self, font="Courier 14", exportselection=False, state='disabled',
                                             background="white", foreground="black", highlightthickness=0)
        self.log_output.pack(fill='both', expand=1, padx=3, pady=3)
        self.selected = False

    def add_message(self, message):
        self.log_output.configure(state="normal")
        self.log_output.insert(tk.END, message)
        self.log_output.see(tk.END)
        self.log_output.configure(state="disabled")

    def mark_as_read(self, *args, **kwargs):
        self.master.tab(self, text="Log")

    def notify(self):
        if not self.selected:
            self.master.tab(self, text="* Log *")

    def tab_selected(self):
        self.selected = True
        self.mark_as_read()

    def tab_deselected(self):
        self.selected = False

    def on_paho_debug_toggle(self):
        if bool(int(self.allow_paho_debug.get())):
            self.log.allow_paho_debug = True
        else:
            self.log.allow_paho_debug = False



"""
MQTTk - Lightweight graphical MQTT client and message analyser

Copyright (C) 2022  Máté Szabó

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

from paho.mqtt.client import MQTTv5, MQTTv31, MQTTv311

COLOURS = ['#9e0505', '#06941b', '#0f05a1', '#999c03', '#048c85', '#5d047a', '#7a3f04',
           '#3b9669', '#b88511', '#1a5e99']

CONNECT = "connected"
DISCONNECT = "disconnected"
QOS_NAMES = {
    "QoS 0": 0,
    "QoS 1": 1,
    "QoS 2": 2
}
DECODER_OPTIONS = [
    "Plain data",
    "JSON pretty formatter",
    "Hex formatter"
]


PROTOCOL_LOOKUP = {
    "3.1": MQTTv31,
    "3.1.1": MQTTv311,
    "5.0": MQTTv5
}

SSL_LIST = ["Disabled", "CA signed server certificate", "CA certificate file", "Self-signed certificate"]

ERROR_CODES = {
    1: "Incorrect protocol version",
    2: "Invalid client identifier",
    3: "Server unavailable",
    4: "Bad username or password",
    5: "Not authorised"
}

EVENT_LEVELS = {
    0: "[i]",
    1: "[w]",
    2: "[E]",
    3: "[X]"
}


MQTT_VERSION_LIST = list(PROTOCOL_LOOKUP.keys())



from functools import partial


def validate_int(d, i, P, s, S, v, V, W):
    try:
        int(S)
    except Exception:
        return False
    return True


def validate_name(name, name_list):
    if name not in name_list:
        return name

    template = name + " {}"
    index = 1
    while template.format(index) in name_list:
        index += 1
    return template.format(index)


def clear_combobox_selection(*_, combobox_instance):
    current_selection = combobox_instance.get()
    combobox_instance.set("")
    combobox_instance.set(current_selection)


def get_clear_combobox_selection_function(combobox_instance):
    return partial(clear_combobox_selection, combobox_instance=combobox_instance)



"""
MQTTk - Lightweight graphical MQTT client and message analyser

Copyright (C) 2022  Máté Szabó

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

import tkinter as tk
import tkinter.ttk as ttk

from mqttk.constants import CONNECT
from mqttk.helpers import get_clear_combobox_selection_function


class HeaderFrame(ttk.Frame):
    def __init__(self, master, app, *args, **kwargs):
        super().__init__(master=master, *args, **kwargs)

        self.connection_selector = ttk.Combobox(self, width=30, exportselection=False)
        self.connection_selector.bind("<<ComboboxSelected>>",
                                      get_clear_combobox_selection_function(self.connection_selector))
        self.connection_selector.pack(side=tk.LEFT, padx=3, pady=3)
        self.connection_selector.configure(state="readonly")
        self.config_window_button = ttk.Button(self, width=10, text="Configure", command=app.spawn_configuration_window)
        self.config_window_button.pack(side=tk.LEFT, expand=False, padx=3, pady=3)
        self.connect_button = ttk.Button(self, width=10, text="Connect", command=app.on_connect_button)
        self.connect_button.pack(side=tk.LEFT, expand=False, padx=3, pady=3)
        self.disconnect_button = ttk.Button(self,
                                            width=10,
                                            text="Disconnect",
                                            state="disabled",
                                            command=app.on_disconnect_button)
        self.disconnect_button.pack(side=tk.LEFT, expand=False, padx=3, pady=3)

        self.connection_indicator = tk.Label(self, text="DISCONNECTED", bg="#ff6b6b")
        self.connection_indicator.pack(side=tk.RIGHT, padx=5, pady=5)
        self.connection_error_notification = ttk.Label(self, foreground='red')
        self.connection_error_notification.pack(side=tk.RIGHT, expand=1, fill='x')

    def interface_toggle(self, connection_state):
        self.connection_selector.configure(state="disabled" if connection_state is CONNECT else "readonly")
        self.config_window_button.configure(state="disabled" if connection_state is CONNECT else "normal")
        self.connect_button.configure(state="disabled" if connection_state is CONNECT else "normal")
        self.disconnect_button.configure(state="normal" if connection_state is CONNECT else "disabled")

    def connection_indicator_toggle(self, connection_state):
        self.connection_indicator.configure(text='CONNECTED' if connection_state == CONNECT else "DISCONNECTED",
                                            bg="#76ff61" if connection_state == CONNECT else "#ff6b6b")



"""
MQTTk - Lightweight graphical MQTT client and message analyser

Copyright (C) 2022  Máté Szabó

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
import json
import os
import traceback
from datetime import datetime
import sys
import time
import csv
from functools import partial
try:
    import tkinter as tk
    import tkinter.ttk as ttk
    from tkinter import messagebox
    from tkinter import filedialog
except ImportError:
    print("Couldn't find a valid installation of tkinter/ttk. "
          "Please make sure you installed the necessary requirements!")
    if sys.platform == "darwin":
        print("Install the python-tk package via homebrew")
    elif sys.platform == "linux":
        print("Install the python-tk or python3-tk package via your operating system's package manager!")
    exit(0)


from mqttk.widgets.subscribe_tab import SubscribeTab
from mqttk.widgets.header_frame import HeaderFrame
from mqttk.widgets.publish_tab import PublishTab
from mqttk.widgets.broker_stats import BrokerStats
from mqttk.constants import CONNECT, DISCONNECT, EVENT_LEVELS
from mqttk.widgets.log_tab import LogTab
from mqttk.widgets.topic_browser import TopicBrowser
from mqttk.widgets.dialogs import AboutDialog, SplashScreen, ConnectionConfigImportExport, SubscribePublishImportExport
from mqttk.widgets.configuration_dialog import ConfigurationWindow
from mqttk.config_handler import ConfigHandler
from mqttk.MQTT_manager import MqttManager
from paho.mqtt.client import MQTT_LOG_ERR, MQTT_LOG_INFO, MQTT_LOG_NOTICE, MQTT_LOG_WARNING


__author__ = "Máté Szabó"
__copyright__ = "Copyright 2022, Máté Szabó"
__credits__ = ["Máté Szabó"]
__license__ = "GPLv3"
__maintainer__ = "Máté Szabó"
__status__ = "Production"


root = tk.Tk()


class PotatoLog:
    def __init__(self):
        self.add_message_callback = None
        self.message_queue = []
        self.config_handler = None
        self.notification_callback = None
        self.allow_paho_debug = False

    def add_message(self, message_level, *args):
        message = "{} - {} ".format(datetime.now().strftime("%Y/%m/%d, %H:%M:%S.%f"), EVENT_LEVELS.get(message_level))
        message += ", ".join([str(x) for x in args])
        message += os.linesep
        if self.add_message_callback is None:
            self.message_queue.append(message)
        else:
            if len(self.message_queue) != 0:
                for queued_message in self.message_queue:
                    self.add_message_callback(queued_message)
                    self.config_handler.add_log_message(queued_message)
                self.message_queue = []
            self.add_message_callback(message)
            self.config_handler.add_log_message(message)
            if 1 < message_level and self.notification_callback is not None:
                self.notification_callback()

    def warning(self, *args):
        self.add_message(1, *args)

    def error(self, *args):
        self.add_message(2, *args)

    def exception(self, *args):
        self.add_message(3, *args)

    def info(self, *args):
        self.add_message(0, *args)

    def on_paho_log(self, _, __, level, buf):
        if level == MQTT_LOG_INFO:
            self.info("[M] " + buf)
        elif level == MQTT_LOG_NOTICE:
            self.info("[M] " + buf)
        elif level == MQTT_LOG_WARNING:
            self.warning("[M] " + buf)
        elif level == MQTT_LOG_ERR:
            self.error("[M] " + buf)
        else:
            if self.allow_paho_debug:
                self.info("[MD]" + buf)


class App:
    def __init__(self, root):
        self.log = PotatoLog()
        self.config_handler = ConfigHandler(self.log)
        self.log.config_handler = self.config_handler

        self.mqtt_manager = None
        self.base64_only = tk.IntVar()
        self.base64_only.set(self.config_handler.get_export_encode_selection())

        root.title("MQTTk")

        # Restore window size and position, if not available or out of bounds, reset to default
        screenwidth = root.winfo_screenwidth()
        screenheight = root.winfo_screenheight()
        saved_geometry = self.config_handler.get_window_geometry()
        out_of_bounds = True
        if saved_geometry is not None:
            out_of_bounds = bool(screenwidth < int(saved_geometry.split("+")[1]) or screenheight < int(
                saved_geometry.split("+")[2]))

        if out_of_bounds:
            width = 1300
            height = 900
            alignstr = '%dx%d+%d+%d' % (width, height, (screenwidth - width) / 2, (screenheight - height) / 2)
            root.geometry(alignstr)
        else:
            root.geometry(saved_geometry)
        root.protocol("WM_DELETE_WINDOW", self.on_destroy)
        root.resizable(width=True, height=True)

        # App icon stuff
        current_dir = os.path.join(os.path.dirname(__file__))
        self.icon_small = tk.PhotoImage(file=os.path.join(current_dir, "mqttk_small.png"))
        self.icon = tk.PhotoImage(file=os.path.join(current_dir, "mqttk.png"))
        self.splash_icon = tk.PhotoImage(file=os.path.join(current_dir, "mqttk_splash.png"))
        root.withdraw()
        splash_screen = SplashScreen(root, self.splash_icon)
        time.sleep(2)
        splash_screen.destroy()
        root.deiconify()
        self.root = root
        self.root.iconphoto(False, self.icon)
        self.root.option_add('*tearOff', tk.FALSE)

        # Some minimal styling stuff
        self.style = ttk.Style()
        if sys.platform == "win32":
            self.style.theme_use('winnative')
        if sys.platform == "darwin":
            self.style.theme_use("default")  # aqua, clam, alt, default, classic
            root.createcommand('tk::mac::ReopenApplication', root.deiconify)
            root.createcommand('tk::mac::ShowHelp', self.on_about_menu)
            # root.createcommand('tk::mac::ShowPreferences', self.show_preferences)  # set preferences menu

        self.style.configure("New.TFrame", background="#b3ffb5")
        self.style.configure("New.TLabel", background="#b3ffb5")
        self.style.configure("Selected.TFrame", background="#96bfff")
        self.style.configure("Selected.TLabel", background="#96bfff")
        self.style.configure("Retained.TLabel", background="#ffeeab")

        # ==================================== Menu bar ===============================================================
        self.menubar = tk.Menu(root, background=self.style.lookup("TLabel", "background"),
                               foreground=self.style.lookup("TLabel", "foreground"))
        self.root.config(menu=self.menubar)

        self.file_menu = tk.Menu(self.menubar, background=self.style.lookup("TLabel", "background"),
                                 foreground=self.style.lookup("TLabel", "foreground"))

        self.import_menu = tk.Menu(self.menubar, background=self.style.lookup("TLabel", "background"),
                                   foreground=self.style.lookup("TLabel", "foreground"))

        self.about_menu = tk.Menu(self.menubar, background=self.style.lookup("TLabel", "background"),
                                  foreground=self.style.lookup("TLabel", "foreground"))

        self.export_menu = tk.Menu(self.menubar, background=self.style.lookup("TLabel", "background"),
                                   foreground=self.style.lookup("TLabel", "foreground"))

        self.export_messages_menu = tk.Menu(self.menubar, background=self.style.lookup("TLabel", "background"),
                                            foreground=self.style.lookup("TLabel", "foreground"))

        self.menubar.add_cascade(menu=self.file_menu, label="File")
        self.file_menu.add_command(label="Exit", command=self.on_exit)

        self.menubar.add_cascade(menu=self.import_menu, label="Import")
        self.import_menu.add_command(label="MQTT.fx config", command=self.import_mqttfx_config)
        self.import_menu.add_command(label="Connection configuration", command=self.import_connection_config)
        self.import_menu.add_command(label="Subscribe/publish content", command=self.import_subscribe_publish)

        self.menubar.add_cascade(menu=self.export_menu, label="Export")
        self.export_menu.add_cascade(menu=self.export_messages_menu, label="Messages")
        self.export_messages_menu.add_command(label="All messages as JSON", command=partial(self.export_messages, format="JSON"))
        self.export_messages_menu.add_command(label="All messages as CSV", command=partial(self.export_messages, format="CSV"))
        self.export_messages_menu.add_separator()
        self.export_messages_menu.add_radiobutton(label="Base64 encode all message payload", value=1, variable=self.base64_only, command=self.save_export_selection)
        self.export_messages_menu.add_radiobutton(label="Base64 encode binary payload only", value=0, variable=self.base64_only, command=self.save_export_selection)
        self.export_messages_menu.add_separator()
        self.export_messages_menu.add_command(label="Current message payload as raw data", command=partial(self.export_messages, format="RAW"))
        self.export_menu.add_command(label="Connection configuration", command=self.export_connection_config)
        self.export_menu.add_command(label="Subscribe/publish content", command=self.export_subscribe_publish)

        self.menubar.add_cascade(menu=self.about_menu, label="Help")
        self.about_menu.add_command(label="About MQTTk", command=self.on_about_menu)

        self.main_window_frame = ttk.Frame(root)
        self.main_window_frame.pack(fill='both', expand=1)

        # ==================================== Header frame ===========================================================
        self.header_frame = HeaderFrame(self.main_window_frame, self, height=35)
        self.header_frame.pack(anchor="w", side=tk.TOP, fill=tk.X, padx=3, pady=3)

        self.tabs = ttk.Notebook(self.main_window_frame)
        self.tabs.pack(anchor="nw", fill="both", expand=True, padx=3, pady=3)

        self.on_config_update()

        # ==================================== Subscribe tab ==========================================================

        self.subscribe_frame = SubscribeTab(self.tabs, self.config_handler, self.log, self.style)
        self.tabs.add(self.subscribe_frame, text="Subscribe")
        self.subscribe_frame.autoscroll_state.set(int(self.config_handler.get_autoscroll()))
        self.subscribe_frame.attempt_to_decompress.set(int(self.config_handler.get_decompress()))
        self.subscribe_frame.message_decoder_selector.set(self.config_handler.get_decoder())

        # ====================================== Publish tab =========================================================

        self.publish_frame = PublishTab(self.tabs, self, self.log, self.style)
        self.tabs.add(self.publish_frame, text="Publish")

        # ====================================== Topic browser tab ====================================================

        self.topic_browser = TopicBrowser(self.tabs, self.config_handler, self.log, root)
        self.tabs.add(self.topic_browser, text="Topic browser")

        # ====================================== Broker stats tab ====================================================

        self.broker_stats = BrokerStats(self.tabs, root, self.log)
        self.tabs.add(self.broker_stats, text="Broker stats")

        # ====================================== Log tab =============================================================

        self.log_tab = LogTab(self.tabs, self.log)
        self.log.notification_callback = self.log_tab.notify
        self.tabs.add(self.log_tab, text="Log")
        self.log.add_message_callback = self.log_tab.add_message
        self.log.info("Logger output live")
        self.tabs.bind("<<NotebookTabChanged>>", self.on_tab_select)

        self.subscribe_frame.interface_toggle(DISCONNECT, None, None)
        self.topic_browser.interface_toggle(DISCONNECT, None, None)
        self.header_frame.interface_toggle(DISCONNECT)
        self.publish_frame.interface_toggle(DISCONNECT)
        self.broker_stats.interface_toggle(DISCONNECT, None)

    def on_client_disconnect(self, notify=None):
        if notify is not None:
            self.header_frame.connection_error_notification["text"] = notify
        self.subscribe_frame.cleanup_subscriptions()
        try:
            self.subscribe_frame.interface_toggle(DISCONNECT, None, None)
            self.topic_browser.interface_toggle(DISCONNECT, None, None)
            self.header_frame.interface_toggle(DISCONNECT)
            self.publish_frame.interface_toggle(DISCONNECT)
            self.broker_stats.interface_toggle(DISCONNECT, None)
            self.header_frame.connection_indicator_toggle(DISCONNECT)
        except Exception as e:
            self.log.exception("Failed to toggle user interface element!", e)
        self.mqtt_manager = None

    def on_client_connect(self):
        self.subscribe_frame.interface_toggle(CONNECT, self.mqtt_manager, self.header_frame.connection_selector.get())
        self.topic_browser.interface_toggle(CONNECT, self.mqtt_manager, self.header_frame.connection_selector.get())
        self.broker_stats.interface_toggle(CONNECT, self.mqtt_manager)
        self.publish_frame.interface_toggle(CONNECT, self.mqtt_manager, self.header_frame.connection_selector.get())
        self.header_frame.connection_indicator_toggle(CONNECT)
        self.subscribe_frame.load_subscription_history()
        self.topic_browser.load_subscription_history()

    def on_connect_button(self):
        if self.header_frame.connection_selector.get() == "":
            return
        self.header_frame.connection_error_notification["text"] = ""
        self.header_frame.interface_toggle(CONNECT)
        self.config_handler.update_last_used_connection(self.header_frame.connection_selector.get())
        try:
            self.mqtt_manager = MqttManager(self.config_handler.get_connection_broker_parameters(self.header_frame.connection_selector.get()),
                                            self.on_client_connect,
                                            self.on_client_disconnect,
                                            self.log)
        except Exception as e:
            self.log.exception("Failed to initialise MQTT client:", e, "\r\n", traceback.format_exc())
            self.header_frame.connection_error_notification["text"] = "Failed to connect, see log for details"
            self.header_frame.interface_toggle(DISCONNECT)
            self.publish_frame.interface_toggle(DISCONNECT)

    def check_disconnect(self):
        if self.mqtt_manager is not None and self.mqtt_manager.disconnect_requested:
            self.mqtt_manager.on_disconnect(0, 0, 0)

    def on_disconnect_button(self):
        self.header_frame.disconnect_button.configure(state="disabled")
        if self.mqtt_manager is not None:
            self.mqtt_manager.disconnect()
            root.after(2000, self.check_disconnect)

    def on_config_update(self):
        connection_profile_list = sorted(self.config_handler.get_connection_profiles())
        if len(connection_profile_list) != 0:
            self.header_frame.connection_selector.configure(values=connection_profile_list)
            if self.config_handler.get_last_used_connection() in connection_profile_list:
                self.header_frame.connection_selector.current(
                    connection_profile_list.index(self.config_handler.get_last_used_connection()))
            else:
                self.header_frame.connection_selector.current(0)

    def spawn_configuration_window(self):
        self.header_frame.connection_error_notification["text"] = ""
        configuration_window = ConfigurationWindow(self.root,
                                                   self.config_handler,
                                                   self.on_config_update,
                                                   self.log,
                                                   self.icon,
                                                   self.header_frame.connection_selector.get())

    def on_about_menu(self):
        about_window = AboutDialog(self.root, self.icon_small, self.style)

    def on_exit(self):
        self.on_disconnect_button()
        self.config_handler.save_window_geometry(self.root.geometry())
        self.config_handler.save_autoscroll(self.subscribe_frame.autoscroll_state.get())
        self.config_handler.save_decompress(self.subscribe_frame.attempt_to_decompress.get())
        self.config_handler.save_decoder(self.subscribe_frame.message_decoder_selector.get())
        root.after(100, root.destroy())
        # root.destroy()

    def on_destroy(self):
        self.on_exit()

    def import_mqttfx_config(self):
        response = messagebox.askquestion("Warning", "This feature is experimental. Would you like to proceed?", )
        if response == "no":
            return
        success = self.config_handler.import_mqttfx_config()
        if success:
            self.on_config_update()

    def export_messages(self, format):

        if self.tabs.tab(self.tabs.select(), "text") not in ("Subscribe"):
            messagebox.showinfo("Info", 'Please engage this operation on the "Subscribe" tab')
            return

        if self.subscribe_frame.message_list_length() == 0:
            messagebox.showinfo("Info", "The message list is empty")
            return

        selected_message_payload = None
        if format == "RAW":
            selected_message_payload = self.subscribe_frame.get_selected_message_payload()
            if selected_message_payload is None:
                messagebox.showinfo("Info", "No message has been selected or message is invalid")
                return

        output_location = filedialog.asksaveasfilename(initialdir=self.config_handler.get_last_used_directory(),
                                                       title="Export {}".format(format),
                                                       defaultextension=format.lower(),
                                                       initialfile="MQTTk_messages_{}".format(time.time()))
        if output_location == "":
            self.log.warning("Empty file name on export message (maybe the cancel button was pressed?")
            return

        self.log.info("Exporting message(s) in {} format to {}".format(format,
                                                                       output_location))

        self.config_handler.save_last_used_directory(output_location)

        try:
            if format == "RAW":
                data = selected_message_payload
                with open(output_location, "wb") as outputfile:
                    outputfile.write(data)

            if format == "JSON":
                messages = []
                for message in self.subscribe_frame.get_messages(bool(int(self.base64_only.get()))):
                    messages.append(message)
                data = json.dumps(messages, indent=2, ensure_ascii=False)
                with open(output_location, "w", encoding="utf-8") as outputfile:
                    outputfile.write(data)

            if format == "CSV":
                with open(output_location, "w", encoding="utf-8") as outputfile:
                    output_writer = csv.writer(outputfile, quoting=csv.QUOTE_MINIMAL, delimiter=',', quotechar='"')
                    header = "timestamp,date,time,subscription pattern,topic,QoS,retained,payload"
                    output_writer.writerow(header)
                    for message in self.subscribe_frame.get_messages(bool(int(self.base64_only.get()))):
                        timestamp = message["timestamp"]
                        datetime_object = datetime.fromtimestamp(timestamp)
                        row = [
                            timestamp,
                            datetime_object.strftime("%Y/%m/%d"),
                            datetime_object.strftime("%H:%M:%S.%f"),
                            message["subscription_pattern"],
                            message["topic"],
                            message["qos"],
                            message["retained"],
                            message["payload"]
                        ]
                        output_writer.writerow(row)

        except Exception as e:
            self.log.exception("Failed to export message data", e, traceback.format_exc())
            messagebox.showerror("Failed to export messages", "Failed to export messages: {} See log for details".format(e))
        else:
            self.log.info("Messages exported successfully")
            messagebox.showinfo("Success", "Messages exported successfully")

    def export_connection_config(self):
        export_dialog = ConnectionConfigImportExport(self.root, self.icon, self.config_handler, self.log, False)

    def import_connection_config(self):
        import_dialog = ConnectionConfigImportExport(self.root, self.icon, self.config_handler, self.log, True)

    def import_subscribe_publish(self):
        import_dialog = SubscribePublishImportExport(self.root, self.icon, self.config_handler, self.log, True)

    def export_subscribe_publish(self):
        export_dialog = SubscribePublishImportExport(self.root, self.icon, self.config_handler, self.log, False)

    def on_tab_select(self, *args, **kwargs):
        if "logtab" in self.tabs.select():
            self.log_tab.tab_selected()
        else:
            self.log_tab.tab_deselected()
            
        # Solves display errors on Mac mini M1 (Monterey) 
        root.after(50, lambda: self.tabs.tab(self.tabs.select(), text=self.tabs.tab(self.tabs.select(), "text")))

    def save_export_selection(self, *args, **kwargs):
        self.config_handler.save_export_encode_selection(int(self.base64_only.get()))


def main():
    app = App(root)
    root.mainloop()


if __name__ == "__main__":
    sys.exit(main())



"""
MQTTk - Lightweight graphical MQTT client and message analyser

Copyright (C) 2022  Máté Szabó

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

import tkinter as tk
import tkinter.ttk as ttk
from tkinter.colorchooser import askcolor
import base64
import json
from os import linesep
import traceback
from functools import partial
import time
from datetime import datetime
import zlib
from bz2 import decompress
from multiprocessing import Lock
from copy import deepcopy

from mqttk.widgets.scroll_frame import ScrollFrame
from mqttk.widgets.scrolled_text import CustomScrolledText
from mqttk.constants import CONNECT, DECODER_OPTIONS, COLOURS
from mqttk.hex_printer import hex_viewer
from mqttk.helpers import get_clear_combobox_selection_function, clear_combobox_selection

ZLIB_TAG0 = chr(0x78)
ZLIB_TAG1 = (chr(0x01), chr(0x5E), chr(0x9C), chr(0xDA))


def decompress_message(message_data):
    try:
        return zlib.decompress(message_data)
    except Exception:
        pass

    try:
        return decompress(message_data)
    except Exception:
        pass

    return message_data


class SubscriptionFrame(ttk.Frame):
    def __init__(self,
                 container,
                 topic,
                 unsubscribe_callback,
                 colour,
                 on_colour_change,
                 mute_callback,
                 *args, **kwargs):
        super().__init__(container, *args, **kwargs)
        self.container = container
        self.topic = topic
        self.colour = colour
        self.unsubscribe_callback = unsubscribe_callback
        self.on_colour_change_callback = on_colour_change
        self.mute_callback = mute_callback
        self.mute_state = False
        self.current_connection = None

        self["relief"] = "groove"
        self["borderwidth"] = 2

        self.topic_frame = ttk.Frame(self)
        self.topic_frame.pack(side=tk.TOP, expand=1, fill='x')
        self.topic_label = ttk.Label(self.topic_frame)
        self.topic_label["text"] = topic
        self.topic_label.pack(side=tk.LEFT, fill="x", expand=1, padx=2, pady=2)

        self.options_frame = ttk.Frame(self)
        self.options_frame.pack(side=tk.BOTTOM, expand=1, fill='x')
        self.unsubscribe_button = ttk.Button(self.options_frame, text="Unsubscribe")
        self.unsubscribe_button.pack(side=tk.RIGHT, padx=2, pady=2)
        self.unsubscribe_button["command"] = self.on_unsubscribe

        self.mute_state_checkbutton = tk.IntVar()
        self.mute_button = ttk.Checkbutton(self.options_frame,
                                           text="Mute",
                                           variable=self.mute_state_checkbutton,
                                           onvalue=1,
                                           offvalue=0)
        self.mute_button.pack(side=tk.RIGHT, padx=4, pady=4)
        self.mute_button['command'] = self.on_mute

        self.colour_picker = ttk.Label(self.options_frame, width=2, background=colour)
        self.colour_picker.bind("<Button-1>", self.on_colour_change)
        self.colour_picker.pack(side=tk.LEFT)

    def on_mute(self):
        self.mute_callback(self.topic, int(self.mute_state_checkbutton.get()))

    def on_unsubscribe(self):
        if self.unsubscribe_callback is not None:
            self.unsubscribe_callback(self.topic)
        self.pack_forget()
        self.destroy()

    def on_colour_change(self, *args, **kwargs):
        colors = askcolor(title="Pick a colour")
        if colors is not None:
            self.colour = colors[1]
            self.colour_picker.configure(background=self.colour)
            self.on_colour_change_callback(self.topic, self.colour)


class SubscribeTab(ttk.Frame):
    def __init__(self, master, config_handler, log, root_style, *args, **kwargs):
        super().__init__(master=master, *args, **kwargs)
        self.config_handler = config_handler
        self.log = log
        self.subscription_frames = {}
        self.color_carousel = -1
        self.current_connection = None
        self.last_connection = None
        self.exporting = False
        # Holds messages and relevant stuff
        # {
        #     "id": {
        #         "topic": "message topic",
        #         "subscription_pattern": "subscription pattern",
        #         "timestamp": "date of reception timestamp or date string whatever",
        #         "qos": "message qos",
        #         "payload": "message content",
        #         "message_list_instance_ref": message list object reference
        #     }
        #
        # }
        self.messages = {}

        self.mute_patterns = []
        self.mqtt_manager = None
        self.message_id_counter = 0

        background_colour = root_style.lookup("TLabel", "background")
        foreground_colour = root_style.lookup("TLabel", "foreground")

        # Subscribe frame
        self.subscribe_bar_frame = ttk.Frame(self, height=1)
        self.subscribe_bar_frame.pack(anchor="nw", side=tk.TOP, fill=tk.X)
        # Subscribe selector combobox
        self.subscribe_selector = ttk.Combobox(self.subscribe_bar_frame, width=30, exportselection=False)
        self.subscribe_selector.bind("<<ComboboxSelected>>",
                                     get_clear_combobox_selection_function(self.subscribe_selector))
        self.subscribe_selector.pack(side=tk.LEFT, padx=3, pady=3)
        self.subscribe_selector["values"] = []
        # Subscribe button
        self.subscribe_button = ttk.Button(self.subscribe_bar_frame, width=10)
        self.subscribe_button.pack(side=tk.LEFT, padx=3, pady=3)
        self.subscribe_button["text"] = "Subscribe"
        self.subscribe_button["command"] = self.add_subscription
        # Flush messages button
        self.flush_messages_button = ttk.Button(self.subscribe_bar_frame, text="Clear messages")
        self.flush_messages_button.pack(side=tk.RIGHT, padx=3)
        self.flush_messages_button["command"] = self.flush_messages
        # Autoscroll checkbox
        self.autoscroll_state = tk.IntVar()
        self.autoscroll_checkbox = ttk.Checkbutton(self.subscribe_bar_frame,
                                                   text="Autoscroll",
                                                   variable=self.autoscroll_state,
                                                   offvalue=0,
                                                   onvalue=1)
        self.autoscroll_checkbox.pack(side=tk.RIGHT, padx=3)

        # Subscribe bottom part frame
        self.subscribe_tab_bottom_frame = ttk.Frame(self)
        self.subscribe_tab_bottom_frame.pack(fill="both", anchor="w", expand=True, padx=3, pady=3)
        # Subscription list paned window
        self.subscription_paned_window = tk.PanedWindow(self.subscribe_tab_bottom_frame,
                                                        orient=tk.HORIZONTAL,
                                                        sashrelief="groove",
                                                        sashwidth=6,
                                                        sashpad=2,
                                                        background=background_colour)
        self.subscription_paned_window.pack(side=tk.LEFT, fill="both", expand=1)
        self.subscriptions_frame = ScrollFrame(self.subscribe_tab_bottom_frame)
        self.subscriptions_frame.pack(fill="y", side=tk.LEFT)
        self.subscription_paned_window.add(self.subscriptions_frame)

        # Incoming message resizable panel
        self.message_paned_window = tk.PanedWindow(self.subscribe_tab_bottom_frame,
                                                   orient=tk.VERTICAL,
                                                   sashrelief="groove",
                                                   sashwidth=6,
                                                   sashpad=2,
                                                   background=background_colour)
        self.message_paned_window.pack(fill='both', padx=3, pady=3, expand=1)
        self.subscription_paned_window.add(self.message_paned_window)

        # Incoming messages listbox
        self.incoming_messages_frame = ttk.Frame(self.subscribe_tab_bottom_frame)
        self.incoming_messages_frame.pack(expand=1, fill='both')

        self.incoming_messages_list = tk.Listbox(self.incoming_messages_frame, selectmode="browse",
                                                 font="Courier 13", background=background_colour)  # TkFixedFont, "Courier 13"
        self.incoming_messages_list.bind("<<ListboxSelect>>", self.on_message_select)

        self.incoming_messages_scrollbar = ttk.Scrollbar(self.incoming_messages_frame,
                                                         orient='vertical',
                                                         command=self.incoming_messages_list.yview)
        self.incoming_messages_list['yscrollcommand'] = self.incoming_messages_scrollbar.set
        self.incoming_messages_scrollbar.pack(side=tk.RIGHT, fill='y')

        self.incoming_messages_scrollbar_h = ttk.Scrollbar(self.incoming_messages_frame,
                                                           orient='horizontal',
                                                           command=self.incoming_messages_list.xview)
        self.incoming_messages_list['xscrollcommand'] = self.incoming_messages_scrollbar_h.set
        self.incoming_messages_scrollbar_h.pack(side=tk.BOTTOM, fill='x')
        self.incoming_messages_list.pack(side=tk.LEFT, fill='both', expand=1)

        self.message_paned_window.add(self.incoming_messages_frame, height=300)

        # Incoming messages scrollable frame
        # self.incoming_messages = ScrollFrame(self.subscribe_tab_bottom_frame)
        # self.incoming_messages.pack()
        # self.message_paned_window.add(self.incoming_messages)

        # Message content frame
        self.message_content_frame = ttk.Frame(self.subscribe_tab_bottom_frame)
        self.message_content_frame.pack(anchor="n", expand=True, fill="both")
        self.message_paned_window.add(self.message_content_frame)

        # Message topic and ID frame
        self.message_topic_and_id_frame = ttk.Frame(self.message_content_frame)
        self.message_topic_and_id_frame.pack(fill="x")
        # Message topic label
        self.message_topic_label = tk.Text(self.message_topic_and_id_frame, height=1, borderwidth=0,
                                           state="disabled", background="white", foreground="black",
                                           exportselection=False)
        self.message_topic_label.pack(side=tk.LEFT, padx=3, pady=3, fill="x", expand=1)
        # Message ID label
        self.message_id_label = ttk.Label(self.message_topic_and_id_frame, width=10)
        self.message_id_label["text"] = "ID"
        self.message_id_label.pack(side=tk.RIGHT, padx=3, pady=3)

        # Message date frame
        self.message_date_and_qos_frame = ttk.Frame(self.message_content_frame)
        self.message_date_and_qos_frame.pack(fill="x")
        # Message date label
        self.message_date_label = ttk.Label(self.message_date_and_qos_frame)
        self.message_date_label["text"] = "DATE"
        self.message_date_label.pack(side=tk.LEFT, padx=3, pady=3)
        # Message QoS label
        self.message_qos_label = ttk.Label(self.message_date_and_qos_frame, width=10)
        self.message_qos_label["text"] = "QOS"
        self.message_qos_label.pack(side=tk.RIGHT, padx=3, pady=3)

        self.attempt_to_decompress = tk.IntVar()
        self.decompress_checkbox = ttk.Checkbutton(self.message_date_and_qos_frame,
                                                   text="Attempt to decompress",
                                                   variable=self.attempt_to_decompress,
                                                   offvalue=0,
                                                   onvalue=1,
                                                   command=self.on_message_select)
        self.decompress_checkbox.pack(side=tk.RIGHT, padx=3)

        # Decoder selector
        self.message_decoder_selector = ttk.Combobox(self.message_date_and_qos_frame,
                                                     width=20,
                                                     state='readonly',
                                                     values=DECODER_OPTIONS,
                                                     exportselection=False)
        self.message_decoder_selector.pack(side=tk.RIGHT, padx=3, pady=3)
        self.message_decoder_selector_label = ttk.Label(self.message_date_and_qos_frame, text="Message decoder")
        self.message_decoder_selector_label.pack(side=tk.RIGHT, padx=3, pady=3)
        self.message_decoder_selector.current(0)
        self.message_decoder_selector.bind("<<ComboboxSelected>>", self.on_decoder_select)

        # Message Payload
        self.message_payload_box = CustomScrolledText(self.message_content_frame,
                                                      exportselection=False,
                                                      background="white",
                                                      foreground="black", highlightthickness=0)
        self.message_payload_box.pack(fill="both", expand=True)
        self.message_payload_box.configure(state="disabled")
        # Message decoder

    def interface_toggle(self, connection_state, mqtt_manager, current_connection):
        # Subscribe tab items
        self.mqtt_manager = mqtt_manager
        if connection_state != CONNECT:
            self.last_connection = self.current_connection
        else:
            if self.last_connection != current_connection:
                self.flush_messages()
        self.current_connection = current_connection
        self.subscribe_button.configure(state="normal" if connection_state is CONNECT else "disabled")
        self.subscribe_selector.configure(state="normal" if connection_state is CONNECT else "disabled")

    def on_decoder_select(self, *args, **kwargs):
        clear_combobox_selection(combobox_instance=self.message_decoder_selector)
        self.on_message_select()
        pass

    def add_message(self, message_title, colour):
        self.incoming_messages_list.insert(tk.END, message_title)
        self.incoming_messages_list.itemconfig(tk.END, fg=colour)
        if bool(self.autoscroll_state.get()):
            self.incoming_messages_list.selection_clear(0, tk.END)
            self.incoming_messages_list.activate(tk.END)
            self.incoming_messages_list.see("end")
            self.incoming_messages_list.selection_set("end", "end")
            self.on_message_select(None)

    def on_message_select(self, *args, **kwargs):
        message_list_id = self.incoming_messages_list.curselection()
        try:
            message_label = self.incoming_messages_list.get(message_list_id)
        except Exception as e:
            self.log.warning("Failed to get message from incoming message list (maybe empty?)", message_list_id)
            message_label = None
        if message_label is None:
            message_id = 0
        else:
            # message_id = int(message_label[-5:])
            message_id = message_list_id[0]

        message_data = self.get_message_details(message_id)
        self.message_topic_label["state"] = "normal"
        self.message_topic_label.delete(1.0, tk.END)
        self.message_topic_label.insert(1.0, message_data.get("topic", ""))
        self.message_topic_label["state"] = "disabled"
        time_string = "{:.6f} - {}".format(round(message_data.get("timestamp", 0), 6),
                                           datetime.fromtimestamp(message_data.get("timestamp", 0)).strftime("%Y/%m/%d, %H:%M:%S.%f"))
        self.message_date_label["text"] = time_string
        self.message_qos_label["text"] = "QoS: {}".format(message_data.get("qos", ""))
        self.message_id_label["text"] = "ID: {}".format(message_id)
        self.message_payload_box.configure(state="normal")
        self.message_payload_box.delete(1.0, tk.END)

        if bool(self.attempt_to_decompress.get()) and 4 < len(message_data.get("payload", "")):
            payload = decompress_message(message_data["payload"])
        else:
            payload = message_data.get("payload", "")

        try:
            payload_decoded = str(payload.decode("utf-8"))
        except Exception:
            payload_decoded = payload
        decoder = self.message_decoder_selector.get()
        if decoder == "JSON pretty formatter":
            try:
                new_message_structure = json.loads(payload_decoded)
            except Exception as e:
                new_message = "        *** FAILED TO LOAD JSON ***{}{}{}{}".format(linesep+linesep,
                                                                                   e,
                                                                                   linesep+linesep,
                                                                                   traceback.format_exc())
            else:
                new_message = json.dumps(new_message_structure, indent=2, ensure_ascii=False)
            self.message_payload_box.insert(1.0, new_message)

        elif decoder == "Hex formatter":
            try:
                data_to_decode = payload_decoded.encode("utf-8")
            except Exception:
                data_to_decode = payload_decoded
            for line in hex_viewer(data_to_decode):
                self.message_payload_box.insert(tk.END, line+linesep)

        else:
            self.message_payload_box.insert(1.0, payload_decoded)
        self.message_payload_box.configure(state="disabled")

    def get_color(self, topic):
        colour = self.config_handler.get_subscription_colour(self.current_connection, topic)
        if colour is not None:
            return colour
        self.color_carousel += 1
        if self.color_carousel > len(COLOURS):
            self.color_carousel = 0
        return COLOURS[self.color_carousel]

    def add_subscription_frame(self, topic, unsubscribe_callback):
        if topic not in self.subscription_frames:
            self.subscription_frames[topic] = SubscriptionFrame(self.subscriptions_frame.viewPort,
                                                                topic,
                                                                unsubscribe_callback,
                                                                self.get_color(topic),
                                                                self.on_colour_change,
                                                                self.topic_mute_callback,
                                                                height=60)
            self.subscription_frames[topic].pack(fill=tk.X, expand=1, padx=2, pady=1)

    def topic_mute_callback(self, topic, mute_state):
        if mute_state and topic not in self.mute_patterns:
            self.mute_patterns.append(topic)
        if not mute_state and topic in self.mute_patterns:
            self.mute_patterns.remove(topic)

    def on_colour_change(self, topic, colour):
        for message_id in list(self.messages.keys()):
            if self.messages[message_id]["subscription_pattern"] == topic:
                self.incoming_messages_list.itemconfig(message_id, fg=colour)
        self.config_handler.add_subscription_history(self.current_connection, topic, colour)

    def add_subscription(self, topic=None):
        if topic is None:
            topic = self.subscribe_selector.get()
        if topic != "" and topic not in self.subscription_frames:
            self.add_subscription_frame(topic, self.on_unsubscribe)
            try:
                callback = partial(self.on_mqtt_message, subscription_pattern=topic)
                callback.__name__ = "MyCallback"  # This is to fix some weird behaviour of the paho client on linux
                self.mqtt_manager.add_subscription(topic_pattern=topic,
                                                   on_message_callback=callback)
            except Exception as e:
                self.log.exception("Failed to subscribe!", e)
                self.subscription_frames[topic].on_unsubscribe()
                return
            # self.add_subscription_frame(topic, self.on_unsubscribe)
            if self.subscribe_selector["values"] == "":
                self.subscribe_selector["values"] = [topic]
            elif topic not in self.subscribe_selector['values']:
                self.subscribe_selector['values'] += (topic,)
            self.config_handler.add_subscription_history(self.current_connection,
                                                         topic,
                                                         self.subscription_frames[topic].colour)

    def add_new_message(self, mqtt_message_object, subscription_pattern):
        timestamp = time.time()
        # Theoretically there will be no race condition here?
        new_message_id = self.message_id_counter
        self.message_id_counter += 1
        simple_time_string = datetime.fromtimestamp(round(timestamp, 3)).strftime("%H:%M:%S.%f")[:-3]
        self.messages[new_message_id] = {
            "topic": mqtt_message_object.topic,
            "payload": mqtt_message_object.payload,
            "qos": mqtt_message_object.qos,
            "subscription_pattern": subscription_pattern,
            "retained": mqtt_message_object.retain,
            "timestamp": timestamp
        }
        message_title = "{} #{:05d} [QoS:{}] [{}] - {}".format(simple_time_string,
                                                               new_message_id,
                                                               mqtt_message_object.qos,
                                                               "R" if mqtt_message_object.retain else " ",
                                                               mqtt_message_object.topic)
        try:
            colour = self.subscription_frames[subscription_pattern].colour
        except Exception as e:
            self.log.warning("Failed to add new message:", e, mqtt_message_object.topic)
        else:
            self.add_message(message_title, colour)

    def load_subscription_history(self):
        self.subscribe_selector.configure(
            values=self.config_handler.get_subscription_history_list(self.current_connection))
        self.subscribe_selector.set(self.config_handler.get_last_subscribe_used(self.current_connection))
        if self.config_handler.get_resubscribe(self.current_connection) == 1:
            topics = self.config_handler.get_resubscribe_topics(self.current_connection)
            for topic in topics:
                self.add_subscription(topic)

    def cleanup_subscriptions(self):
        current_subscriptions = []
        for topic in list(self.subscription_frames.keys()):
            current_subscriptions.append(topic)
            self.subscription_frames[topic].pack_forget()
            self.subscription_frames[topic].destroy()
        if self.config_handler.get_resubscribe(self.current_connection) == 1:
            self.config_handler.save_resubscribe_topics(self.current_connection, current_subscriptions)
        self.subscription_frames = {}

    def on_mqtt_message(self, _, __, msg, subscription_pattern):
        if self.exporting:
            return
        if subscription_pattern in self.mute_patterns:
            return
        self.add_new_message(mqtt_message_object=msg,
                             subscription_pattern=subscription_pattern)

    def get_message_details(self, message_id):
        return self.messages.get(message_id, {})

    def on_unsubscribe(self, topic):
        self.subscription_frames.pop(topic, None)
        try:
            self.mqtt_manager.unsubscribe(topic)
        except Exception as e:
            self.log.warning("Failed to unsubscribe", topic, "maybe a failed subscription?")

    def flush_messages(self):
        self.message_id_counter = 0
        self.incoming_messages_list.delete(0, "end")
        self.messages = {}
        self.on_message_select()

    def message_list_length(self):
        return len(self.messages)

    def get_selected_message_payload(self):
        try:
            message_list_id = self.incoming_messages_list.curselection()
            message_id = message_list_id[0]
            message_data = self.get_message_details(message_id)["payload"]
        except Exception as e:
            return None
        return message_data

    def get_messages(self, base64_only):
        self.exporting = True
        for message in self.messages.values():
            message_to_export = deepcopy(message)
            if base64_only:
                message_to_export["payload"] = base64.b64encode(message_to_export["payload"]).decode("utf-8")
                yield message_to_export
                continue
            try:
                message_to_export["payload"] = message_to_export["payload"].decode("utf-8")
            except Exception:
                message_to_export["payload"] = base64.b64encode(message_to_export["payload"]).decode("utf-8")
            yield message_to_export
        self.exporting = False



"""
MQTTk - Lightweight graphical MQTT client and message analyser

Copyright (C) 2022  Máté Szabó

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

try:
    import xmltodict
    XMLTODICT = True
except ImportError:
    XMLTODICT = False


def validate(value):
    if value is not None:
        return value
    return ""


def parse_mqttfx_xml(mqttfx_config_file):
    try:
        with open(mqttfx_config_file, 'r') as configfile:
            mqttfx_config_xml = configfile.read()

        mqttfx_config_dict = xmltodict.parse(mqttfx_config_xml)
    except Exception as e:
        return True, "Failed to parse MQTT.fx configuration XML: {}".format(e)
    return False, mqttfx_config_dict


def parse_mqttfx_config(mqttfx_config_dict, mqttk_configuration):

    connection_profiles = mqttfx_config_dict.get("configuration", {}).get(
        "connectionProfiles", {}).get("connectionProfile", None)

    if connection_profiles is None:
        return True, "Unexpected MQTT.fx configuration format"

    if "connections" not in mqttk_configuration:
        mqttk_configuration["connections"] = {}

    for connection_profile in connection_profiles:
        profile_name = "MQTT.fx {}".format(connection_profile["profileName"])
        mqttk_configuration["connections"][profile_name] = {
                "connection_parameters": {},
                "subscriptions": {},
                "publish_topics": [],
                "stored_publishes": {}
            }

        mqttk_configuration["connections"][profile_name]["connection_parameters"]["broker_addr"] = validate(connection_profile["brokerAddress"])
        mqttk_configuration["connections"][profile_name]["connection_parameters"]["broker_port"] = validate(connection_profile["brokerPort"])
        mqttk_configuration["connections"][profile_name]["connection_parameters"]["client_id"] = validate(connection_profile["connectionOptions"]["clientId"])
        mqttk_configuration["connections"][profile_name]["connection_parameters"]["user"] = validate(connection_profile["connectionOptions"]["userName"])
        mqttk_configuration["connections"][profile_name]["connection_parameters"]["pass"] = validate(connection_profile["connectionOptions"]["password"])
        mqttk_configuration["connections"][profile_name]["connection_parameters"]["timeout"] = validate(connection_profile["connectionOptions"]["connectionTimeout"])
        mqttk_configuration["connections"][profile_name]["connection_parameters"]["keepalive"] = validate(connection_profile["connectionOptions"]["keepAliveInterval"])
        mqttk_configuration["connections"][profile_name]["connection_parameters"]["mqtt_version"] = validate(connection_profile["connectionOptions"]["mqttVersion"])
        mqttk_configuration["connections"][profile_name]["connection_parameters"]["ssl"] = "Disabled"
        mqttk_configuration["connections"][profile_name]["connection_parameters"]["ca_file"] = validate(connection_profile["connectionOptions"]["caFile"])
        mqttk_configuration["connections"][profile_name]["connection_parameters"]["cl_cert"] = validate(connection_profile["connectionOptions"]["clientCertificateFile"])
        mqttk_configuration["connections"][profile_name]["connection_parameters"]["cl_key"] = validate(connection_profile["connectionOptions"]["clientKeyFile"])
        # TODO add dialog to check ssl configuration

        if connection_profile["preDefinedMessages"] is not None:
            if type(connection_profile["preDefinedMessages"]["message"]) is list:
                for predefined_message in connection_profile["preDefinedMessages"]["message"]:
                    mqttk_configuration["connections"][profile_name]["stored_publishes"][predefined_message['name']] = {
                        "topic": predefined_message['topic']["name"],
                        "qos": int(predefined_message['qos']),
                        "retained": bool(predefined_message['retained']),
                        "payload": predefined_message['payload']
                    }

        if connection_profile["recentPublishTopics"] is not None:
            if type(connection_profile["recentPublishTopics"]["topic"]) is list:
                for topic in connection_profile["recentPublishTopics"]["topic"]:
                    if "#" in topic or '+' in topic or '%' in topic:
                        continue
                    mqttk_configuration["connections"][profile_name]["publish_topics"].append(topic)

        if connection_profile["recentSubscriptionTopics"] is not None:
            if type(connection_profile["recentSubscriptionTopics"]["topic"]) is list:
                for subscription_topic in connection_profile["recentSubscriptionTopics"]["topic"]:
                    mqttk_configuration["connections"][profile_name]["subscriptions"][subscription_topic["name"]] = {
                        "colour": subscription_topic["color"].lower()
                    }

    return False, mqttk_configuration



"""
MQTTk - Lightweight graphical MQTT client and message analyser

Copyright (C) 2022  Máté Szabó

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

import tkinter as tk
import tkinter.ttk as ttk


class CustomScrolledText(tk.Text):
    def __init__(self, master=None, **kw):
        self.frame = ttk.Frame(master)
        self.vbar = ttk.Scrollbar(self.frame)
        self.vbar.pack(side=tk.RIGHT, fill=tk.Y)

        kw.update({'yscrollcommand': self.vbar.set})
        tk.Text.__init__(self, self.frame, **kw)
        self.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        self.vbar['command'] = self.yview

        # Copy geometry methods of self.frame without overriding Text
        # methods -- hack!
        text_meths = vars(tk.Text).keys()
        methods = vars(tk.Pack).keys() | vars(tk.Grid).keys() | vars(tk.Place).keys()
        methods = methods.difference(text_meths)

        for m in methods:
            if m[0] != '_' and m != 'config' and m != 'configure':
                setattr(self, m, getattr(self.frame, m))

    def __str__(self):
        return str(self.frame)



import ssl

import paho.mqtt.client as mqtt
from mqttk.constants import PROTOCOL_LOOKUP, SSL_LIST, ERROR_CODES
from uuid import uuid4


class MqttManager:
    def __init__(self, connection_configuration, on_connect_callback, on_disconnect_callback, logger):
        if not connection_configuration:
            raise Exception("Invalid connection parameters, configuration empty")
        self.on_connect_callback = on_connect_callback
        self.on_disconnect_callback = on_disconnect_callback
        self.log = logger
        self.disconnect_requested = False

        autogen = connection_configuration.get("client_id_autogen", 0)
        if autogen == 1:
            self.client_id = str(uuid4()).replace("-", "")
        else:
            self.client_id = connection_configuration["client_id"]

        try:
            self.client = mqtt.Client(self.client_id,
                                      clean_session=True,
                                      userdata=None,
                                      protocol=PROTOCOL_LOOKUP.get(connection_configuration["mqtt_version"], mqtt.MQTTv311),
                                      transport="tcp")
        except ValueError:
            self.client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION1,
                                      self.client_id,
                                      clean_session=True,
                                      userdata=None,
                                      protocol=PROTOCOL_LOOKUP.get(connection_configuration["mqtt_version"], mqtt.MQTTv311),
                                      transport="tcp")

        self.client.on_log = self.log.on_paho_log

        if connection_configuration.get("user", "") != "":
            self.client.username_pw_set(username=connection_configuration["user"],
                                        password=connection_configuration.get("pass", ""))

        ssl_config = connection_configuration.get("ssl", None)
        if ssl_config is not None and ssl_config in SSL_LIST and ssl_config != "Disabled":
            if ssl_config.startswith("CA signed"):
                self.client.tls_set(ca_certs=None,
                                    certfile=None,
                                    keyfile=None,
                                    cert_reqs=ssl.CERT_REQUIRED,
                                    tls_version=ssl.PROTOCOL_TLS,
                                    ciphers=None,
                                    keyfile_password=None)
            elif ssl_config.startswith("CA certificate"):
                self.client.tls_set(ca_certs=connection_configuration.get("ca_file", ""),
                                    certfile=None,
                                    keyfile=None,
                                    cert_reqs=ssl.CERT_REQUIRED,
                                    tls_version=ssl.PROTOCOL_TLS,
                                    ciphers=None,
                                    keyfile_password=None)
            elif ssl_config.startswith("Self-signed certificate"):
                self.client.tls_set(ca_certs=connection_configuration.get("ca_file", ""),
                                    certfile=connection_configuration.get("cl_cert", ""),
                                    keyfile=connection_configuration.get("cl_key", ""),
                                    cert_reqs=ssl.CERT_REQUIRED,
                                    tls_version=ssl.PROTOCOL_TLS,
                                    ciphers=None,
                                    keyfile_password=None)

        self.client.on_connect = self.on_connect
        self.client.on_disconnect = self.on_disconnect
        self.client.connect(host=connection_configuration.get("broker_addr", ""),
                            port=int(connection_configuration.get("broker_port", "")),
                            keepalive=int(connection_configuration.get("keepalive", 60)))
        self.client.loop_start()
        self.log.info("Paho MQTT client manager initialised")

    def on_connect(self, _, __, ___, rc):
        if rc == 0:
            self.log.info("Paho MQTT Client successfully connected, client ID: {}".format(self.client_id))
            self.client.loop_start()
            self.on_connect_callback()
        else:
            self.log.error("Bad connection, returned code: {}".format(rc))
            self.on_disconnect_callback(notify="Failed to connect: {}".format(ERROR_CODES.get(rc, "Unknown error {}".format(rc))))

    def on_disconnect(self, _, __, rc):
        self.log.info("Paho MQTT client disconnected")
        self.client.loop_stop()
        if rc != 0:
            try:
                self.on_disconnect_callback(notify="Disconnected, return code: {}".format(rc))
            except Exception as e:
                self.log.exception("Failed to reconfigure interface for disconnect: {}".format(e))
        self.on_disconnect_callback()

    def disconnect(self):
        if self.client.is_connected():
            self.log.info("Paho MQTT client manager instructed to disconnect")
            self.client.disconnect()
        else:
            try:
                self.client.disconnect()
            except Exception:
                pass
            self.on_disconnect(0, 0, 0)
        self.disconnect_requested = True

    def add_subscription(self, topic_pattern, on_message_callback):
        self.log.info("MQTT client manager adding subscription", topic_pattern)
        self.client.subscribe(topic_pattern)
        self.client.message_callback_add(topic_pattern, on_message_callback)

    def unsubscribe(self, topic_filter):
        self.log.info("MQTT client manager unsubscribing", topic_filter)
        self.client.unsubscribe(topic_filter)
        self.client.message_callback_remove(topic_filter)

    def publish(self, topic, payload, qos, retained):
        self.log.info("Publish", topic)
        self.client.publish(topic, payload, qos, retained)



"""
MQTTk - Lightweight graphical MQTT client and message analyser

Copyright (C) 2022  Máté Szabó

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

import tkinter as tk
import tkinter.ttk as ttk
from mqttk.widgets.scrolled_text import CustomScrolledText
from mqttk.widgets.scroll_frame import ScrollFrame
from mqttk.widgets.dialogs import PublishNameDialog
from mqttk.helpers import get_clear_combobox_selection_function
import base64
from mqttk.constants import QOS_NAMES, CONNECT, DISCONNECT


class PublishHistoryFrame(ttk.Frame):
    def __init__(self,
                 master,
                 name,
                 config,
                 publish_callback,
                 delete_callback,
                 on_select_callback,
                 on_edit_callback,
                 *args, **kwargs):
        super().__init__(master, *args, **kwargs)

        self.name = name
        self.configuration = config
        self.publish_callback = publish_callback
        self.delete_callback = delete_callback
        self.on_select_callback = on_select_callback
        self.on_edit_callback = on_edit_callback
        self.mqtt_manager = None

        self["relief"] = "groove"
        self["borderwidth"] = 2
        self.bind("<Button-1>", self.on_select)

        self.name_label = ttk.Label(self, text=name, justify="left")
        self.name_label.pack(expand=1, fill="x", side=tk.TOP, padx=3, pady=6)
        self.name_label.bind("<Button-1>", self.on_select)

        self.publish_history_actions = ttk.Frame(self)
        self.publish_history_actions.pack(side=tk.TOP, fill='x', padx=3, pady=3)
        self.publish_history_actions.bind("<Button-1>", self.on_select)

        self.publish_button = ttk.Button(self.publish_history_actions, text="Publish")
        self.publish_button["command"] = self.on_publish_button
        self.publish_button.pack(side=tk.RIGHT, padx=3, pady=3)

        self.edit_button = ttk.Button(self.publish_history_actions, text="Rename")
        self.edit_button["command"] = self.on_edit_button
        self.edit_button.pack(side=tk.LEFT, padx=3, pady=3)

        self.delete_button = ttk.Button(self.publish_history_actions, text="Delete")
        self.delete_button["command"] = self.on_delete_button
        self.delete_button.pack(side=tk.LEFT, padx=3, pady=3)

    def on_select(self, *args, **kwargs):
        self.publish_history_actions.configure(style="Selected.TFrame")
        self.configure(style="Selected.TFrame")
        self.name_label.configure(style="Selected.TLabel")
        self.on_select_callback(self)

    def on_unselect(self, *args, **kwargs):
        self.publish_history_actions.configure(style="TFrame")
        self.configure(style="TFrame")
        self.name_label.configure(style="TLabel")

    def on_delete_button(self):
        self.delete_callback(self.name)

    def on_edit_button(self):
        self.on_select()
        self.on_edit_callback()

    def on_publish_button(self):
        self.publish_callback(self.configuration["topic"],
                              base64.b64decode(self.configuration["payload"]).decode("utf-8"),
                              self.configuration["qos"],
                              self.configuration["retained"])


class PublishTab(ttk.Frame):
    def __init__(self, master, app, log, root_style, *args, **kwargs):
        super().__init__(master=master, *args, **kwargs)

        background_colour = root_style.lookup("TLabel", "background")

        self.app_root = app.root
        self.config_handler = app.config_handler
        self.current_connection = None
        self.mqtt_manager = None
        self.topic_history = []
        self.log = log

        self.current_publish_history_selected = None  # Reference to the selected PublishHistoryFrame
        self.publish_history_frames = {}  # publish history name to object reference
        self.selected_history_unselect_callback = None

        self.publish_paned_window = tk.PanedWindow(self,
                                                   orient=tk.HORIZONTAL,
                                                   sashrelief="groove",
                                                   sashwidth=6,
                                                   sashpad=2,
                                                   background=background_colour)
        self.publish_paned_window.pack(fill='both', expand=1, side=tk.LEFT)
        self.saved_publishes = ScrollFrame(self)
        self.saved_publishes.pack(fill="y", expand=1, side=tk.LEFT)
        self.publish_paned_window.add(self.saved_publishes, width=350)

        self.publish_interface = ttk.Frame(self)
        self.publish_interface.pack(fill='both', expand=1)

        self.publish_interface_actions = ttk.Frame(self.publish_interface)
        self.publish_interface_actions.pack(fill='x', side=tk.TOP)
        self.publish_topic_selector = ttk.Combobox(self.publish_interface_actions, width=40)
        self.publish_topic_selector.bind("<<ComboboxSelected>>",
                                         get_clear_combobox_selection_function(self.publish_topic_selector))
        self.publish_topic_selector.pack(side=tk.LEFT, padx=4, pady=4)

        self.publish_button = ttk.Button(self.publish_interface_actions, text="Publish")
        self.publish_button['command'] = self.on_publish_button
        self.publish_button.pack(side=tk.LEFT, padx=2, pady=4)

        self.save_publish_button = ttk.Button(self.publish_interface_actions, text="Save")
        self.save_publish_button.pack(side=tk.LEFT, padx=4, pady=4)
        self.save_publish_button["command"] = self.on_publish_save

        self.retained_state_var = tk.IntVar()
        self.retained_checkbox = ttk.Checkbutton(self.publish_interface_actions,
                                                 text="Retained",
                                                 onvalue=1,
                                                 offvalue=0,
                                                 variable=self.retained_state_var)
        self.retained_checkbox.pack(side=tk.RIGHT, pady=4, padx=2)

        self.qos_selector = ttk.Combobox(self.publish_interface_actions,
                                         exportselection=False,
                                         width=7,
                                         values=list(QOS_NAMES.keys()))
        self.qos_selector.bind("<<ComboboxSelected>>",
                               get_clear_combobox_selection_function(self.qos_selector))
        self.qos_selector.current(0)
        self.qos_selector.pack(side=tk.RIGHT, pady=4, padx=2)

        self.payload_editor = CustomScrolledText(self.publish_interface,
                                                 font="Courier 13",
                                                 background="white",
                                                 foreground="black", highlightthickness=0)
        self.payload_editor.pack(fill="both", expand=1, side=tk.BOTTOM)
        self.publish_paned_window.add(self.publish_interface)

    def on_publish_history_delete(self, name):
        self.publish_history_frames[name].pack_forget()
        self.publish_history_frames[name].destroy()
        self.config_handler.delete_publish_history_item(self.current_connection, name)

    def publish_message(self, topic, payload, qos, retained):
        if topic not in self.topic_history:
            self.config_handler.save_publish_topic_history_item(self.current_connection, topic)
        try:
            if self.mqtt_manager is not None:
                self.mqtt_manager.publish(topic, payload, qos, retained)
        except Exception as e:
            self.log.exception("Failed to publish!", e, topic, payload, qos, retained)

    def on_publish_button(self, *args, **kwargs):
        if self.publish_topic_selector.get() != "":
            payload = self.payload_editor.get(1.0, tk.END)
            # Remove stupid fucking newline that gets added to the bloody text widget for no reason
            if payload[-1] == '\n':
                payload = payload[0:-1]
            self.publish_message(self.publish_topic_selector.get(),
                                 payload,
                                 QOS_NAMES.get(self.qos_selector.get(), 0),
                                 bool(self.retained_state_var.get()))
            new = self.config_handler.save_publish_topic_history_item(self.current_connection,
                                                                      self.publish_topic_selector.get())
            if new:
                if len(self.config_handler.get_publish_topic_history(self.current_connection)) > 1:
                    self.publish_topic_selector['values'] += (self.publish_topic_selector.get(),)
                else:
                    self.publish_topic_selector['values'] = (self.publish_topic_selector.get(),)

    def on_publish_save(self, *args, **kwargs):
        if self.publish_topic_selector.get() == "":
            return
        current_name = ""
        if self.current_publish_history_selected is not None:
            current_name = self.current_publish_history_selected.name
        name_entry_window = PublishNameDialog(self.app_root, current_name, self.save_new_name_callback)
        name_entry_window.transient(self.app_root)
        name_entry_window.wait_visibility()
        name_entry_window.grab_set()
        name_entry_window.wait_window()

    def on_new_name_rename(self, new_name):
        if self.current_publish_history_selected.name == new_name:
            return
        self.publish_history_frames[new_name] = self.current_publish_history_selected
        self.publish_history_frames.pop(self.current_publish_history_selected.name)
        self.config_handler.save_publish_history_item(self.current_connection,
                                                      new_name,
                                                      self.current_publish_history_selected.configuration)
        self.config_handler.delete_publish_history_item(self.current_connection,
                                                        self.current_publish_history_selected.name)
        self.current_publish_history_selected.name = new_name
        self.current_publish_history_selected.name_label["text"] = new_name

    def on_rename_callback(self):
        current_name = self.current_publish_history_selected.name
        name_entry_window = PublishNameDialog(self.app_root, current_name, self.on_new_name_rename)
        name_entry_window.transient(self.app_root)
        name_entry_window.wait_visibility()
        name_entry_window.grab_set()
        name_entry_window.wait_window()

    def save_new_name_callback(self, new_name):
        new_config = {
            "topic": self.publish_topic_selector.get(),
            "qos": QOS_NAMES.get(self.qos_selector.get(), 0),
            "retained": bool(self.retained_state_var.get()),
            "payload": base64.b64encode(self.payload_editor.get(1.0, tk.END).encode("utf-8")).decode("utf-8")
        }
        self.config_handler.save_publish_history_item(self.current_connection, new_name, new_config)
        if self.current_publish_history_selected is None or self.current_publish_history_selected.name != new_name:
            self.add_new_publish_history_item(new_name, new_config)
            self.publish_history_frames[new_name].on_select()
        else:
            self.current_publish_history_selected.configuration = new_config

    def add_new_publish_history_item(self, name, config):
        self.publish_history_frames[name] = PublishHistoryFrame(self.saved_publishes.viewPort,
                                                                name,
                                                                config,
                                                                self.publish_message,
                                                                self.on_publish_history_delete,
                                                                self.on_publish_history_select,
                                                                self.on_rename_callback)
        self.publish_history_frames[name].pack(fill=tk.X, expand=1, padx=2, pady=1)

    def load_publish_and_topic_history(self, current_connection):
        self.current_connection = current_connection
        self.publish_topic_selector.set(self.config_handler.get_last_publish_topic(self.current_connection))
        publish_history = self.config_handler.get_publish_history(current_connection)
        for name, config in publish_history.items():
            self.add_new_publish_history_item(name, config)
        self.topic_history = self.config_handler.get_publish_topic_history(current_connection)
        self.publish_topic_selector.configure(values=self.topic_history)

    def on_publish_history_select(self, history_item):
        if self.selected_history_unselect_callback is not None and history_item.name != self.current_publish_history_selected.name:
            try:
                self.selected_history_unselect_callback()
            except Exception as e:
                self.log.warning("Failed to deselect item, maybe no longer present?", e)
        self.selected_history_unselect_callback = history_item.on_unselect
        self.publish_topic_selector.set(history_item.configuration["topic"])
        self.qos_selector.current(int(history_item.configuration["qos"]))
        self.retained_state_var.set(history_item.configuration["retained"])
        self.payload_editor.delete(1.0, tk.END)
        self.payload_editor.insert(1.0, base64.b64decode(history_item.configuration["payload"]).decode("utf-8"))
        self.current_publish_history_selected = history_item

    def interface_toggle(self, connection_state, mqtt_manager=None, current_connection=None):
        self.mqtt_manager = mqtt_manager
        if connection_state == CONNECT:
            self.load_publish_and_topic_history(current_connection)
        if connection_state == DISCONNECT:
            for name, publish_history_element in self.publish_history_frames.items():
                publish_history_element.pack_forget()
                publish_history_element.destroy()
            self.publish_history_frames = {}
            self.payload_editor.delete(1.0, tk.END)
            self.publish_topic_selector.configure(values=[])
            self.publish_topic_selector.set("")

        self.publish_button.configure(state="normal" if connection_state is CONNECT else "disabled")
        self.save_publish_button.configure(state="normal" if connection_state is CONNECT else "disabled")
        self.retained_checkbox.configure(state="normal" if connection_state is CONNECT else "disabled")
        self.qos_selector.configure(state="readonly" if connection_state is CONNECT else "disabled")
        self.publish_topic_selector.configure(state="normal" if connection_state is CONNECT else "disabled")
        self.payload_editor.configure(state="normal" if connection_state is CONNECT else "disabled")



from setuptools import setup, find_packages
from mqttk import __version__

url = "https://github.com/matesh/mqttk"
readme = open('README.md').read()

setup(
    name="mqttk",
    packages=find_packages(),
    version=__version__,
    long_description=readme,
    long_description_content_type='text/markdown',
    include_package_data=True,
    author="Mate Szabo",
    author_email="mate@mateszabo.com",
    maintainer='Mate Szabo',
    maintainer_email='mate@mateszabo.com',
    description="A lightweight MQTT client GUI written in pure python",
    url=url,
    python_requires=">=3.7",
    install_requires=[
        "paho-mqtt",
        "xmltodict"
    ],
    entry_points={
        'gui_scripts': ['mqttk=mqttk.__main__:main'],
        'console_scripts': ['mqttk-console=mqttk.__main__:main']
    },
    package_data={
        'mqttk': ['*.png']
    },
    download_url="{}/releases".format(url),
    license="GPLv3"
)
